# Bob build tool
# Copyright (C) 2016  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

def getBobInputHash():
    from .utils import hashDirectory
    import os
    # we need the source hash to invalidate the cache in case of source code changes.
    # therefore it's enough to hash the pym directory without the entries of cmds-Dir
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')
    return hashDirectory(root, ignoreDirs=['__pycache__', 'cmds'])

# First try to see if we're running a development version. If we do we take the
# version from git and make sure everything is up-to-date. Otherwise Bob was
# installed via pip and we can import the installed version.
try:
    from .develop.version import BOB_VERSION
    try:
        BOB_INPUT_HASH = getBobInputHash()
    except KeyboardInterrupt:
        sys.exit(1)
except ImportError:
    from .version import version as BOB_VERSION
    BOB_INPUT_HASH = BOB_VERSION.encode("utf-8")

# global debug switches
DEBUG = {
    'ngd' :  False,     # no-global-defaults
    'pkgck' : False,    # package-calculation-checks
    'prof' : False,     # profiling
    'shl' : False,      # shell-trap
}

# interactive debug shell
def __debugTap(sig, frame):
    import code, traceback

    """Interrupt running process, and provide a python prompt for
    interactive debugging."""
    d={'_frame':frame}         # Allow access to frame object.
    d.update(frame.f_globals)  # Unless shadowed by global
    d.update(frame.f_locals)

    i = code.InteractiveConsole(d)
    message  = "Signal received : entering python shell.\nTraceback:\n"
    message += ''.join(traceback.format_stack(frame))
    i.interact(message)

def _enableDebug(enabled):
    global DEBUG

    for e in enabled.split(','):
        e = e.strip()
        if e in DEBUG:
            DEBUG[e] = True
        else:
            print("Invalid debug flag:", e, file=sys.stderr)
            sys.exit(2)

    if DEBUG['shl']:
        import signal
        signal.signal(signal.SIGUSR1, __debugTap)



# Bob build tool
# Copyright (C) 2016-2018  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ...archive import getArchiver
from ...errors import BuildError
from ...input import RecipeSet
from ...tty import setVerbosity, setTui
from ...utils import copyTree, processDefines, EventLoopWrapper
import argparse
import datetime
import os
import subprocess
import sys
import time

from .state import DevelopDirOracle
from .builder import LocalBuilder

def runHook(recipes, hook, args):
    hookCmd = recipes.getBuildHook(hook)
    ret = True
    if hookCmd:
        try:
            hookCmd = os.path.expanduser(hookCmd)
            ret = subprocess.call([hookCmd] + args) == 0
        except OSError as e:
            raise BuildError(hook + ": cannot run '" + hookCmd + ": " + str(e))

    return ret

def commonBuildDevelop(parser, argv, bobRoot, develop):
    def _downloadArgument(arg):
        if arg.startswith('packages=') or arg in ['yes', 'no', 'deps', 'forced', 'forced-deps', 'forced-fallback']:
            return arg
        raise argparse.ArgumentTypeError("{} invalid.".format(arg))

    parser.add_argument('packages', metavar='PACKAGE', type=str, nargs='+',
        help="(Sub-)package to build")
    parser.add_argument('--destination', metavar="DEST", default=None,
        help="Destination of build result (will be overwritten!)")
    parser.add_argument('-j', '--jobs', default=None, type=int, nargs='?', const=...,
        help="Specifies  the  number of jobs to run simultaneously.")
    parser.add_argument('-k', '--keep-going', default=None, action='store_true',
        help="Continue  as much as possible after an error.")
    parser.add_argument('-f', '--force', default=None, action='store_true',
        help="Force execution of all build steps")
    parser.add_argument('-n', '--no-deps', default=None, action='store_true',
        help="Don't build dependencies")
    parser.add_argument('-p', '--with-provided', dest='build_provided', default=None, action='store_true',
        help="Build provided dependencies")
    parser.add_argument('--without-provided', dest='build_provided', default=None, action='store_false',
        help="Build without provided dependencies")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-A', '--no-audit', dest='audit', default=None,
        action='store_false',
        help="Don't generate audit trail for build results")
    group.add_argument('--audit', dest='audit', action='store_true',
        help="Generate audit trail (default)")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-b', '--build-only', dest='build_mode', default=None,
        action='store_const', const='build-only',
        help="Don't checkout, just build and package")
    group.add_argument('-B', '--checkout-only', dest='build_mode',
        action='store_const', const='checkout-only',
        help="Don't build, just check out sources")
    group.add_argument('--normal', dest='build_mode',
        action='store_const', const='normal',
        help="Checkout, build and package")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--clean', action='store_true', default=None,
        help="Do clean builds (clear build directory)")
    group.add_argument('--incremental', action='store_false', dest='clean',
        help="Reuse build directory for incremental builds")
    parser.add_argument('--always-checkout', default=[], action='append', metavar="RE",
        help="Regex pattern of packages that should always be checked out")
    parser.add_argument('--resume', default=False, action='store_true',
        help="Resume build where it was previously interrupted")
    parser.add_argument('-q', '--quiet', default=0, action='count',
        help="Decrease verbosity (may be specified multiple times)")
    parser.add_argument('-v', '--verbose', default=0, action='count',
        help="Increase verbosity (may be specified multiple times)")
    parser.add_argument('--no-logfiles', default=None, action='store_true',
        help="Disable logFile generation.")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    parser.add_argument('-e', dest="white_list", default=[], action='append', metavar="NAME",
        help="Preserve environment variable")
    parser.add_argument('-E', dest="preserve_env", default=False, action='store_true',
        help="Preserve whole environment")
    parser.add_argument('--upload', default=None, action='store_true',
        help="Upload to binary archive")
    parser.add_argument('--link-deps', default=None, help="Add linked dependencies to workspace paths",
        dest='link_deps', action='store_true')
    parser.add_argument('--no-link-deps', default=None, help="Do not add linked dependencies to workspace paths",
        dest='link_deps', action='store_false')
    parser.add_argument('--download', metavar="MODE", default=None,
        help="Download from binary archive (yes, no, deps, forced, forced-deps, forced-fallback, packages=<packages>",
        type=_downloadArgument)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=None,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")
    parser.add_argument('--clean-checkout', action='store_true', default=None, dest='clean_checkout',
        help="Do a clean checkout if SCM state is dirty.")
    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    startTime = time.time()

    with EventLoopWrapper() as loop:
        recipes = RecipeSet()
        recipes.defineHook('releaseNameFormatter', LocalBuilder.releaseNameFormatter)
        recipes.defineHook('developNameFormatter', LocalBuilder.developNameFormatter)
        recipes.defineHook('developNamePersister', None)
        recipes.setConfigFiles(args.configFile)
        recipes.parse()

        # if arguments are not passed on cmdline use them from default.yaml or set to default yalue
        if develop:
            cfg = recipes.getCommandConfig().get('dev', {})
        else:
            cfg = recipes.getCommandConfig().get('build', {})

        defaults = {
                'destination' : '',
                'force' : False,
                'no_deps' : False,
                'build_mode' : 'normal',
                'clean' : not develop,
                'upload' : False,
                'download' : "deps" if develop else "yes",
                'sandbox' : not develop,
                'clean_checkout' : False,
                'no_logfiles' : False,
                'link_deps' : True,
                'jobs' : 1,
                'keep_going' : False,
                'audit' : True,
            }

        for a in vars(args):
            if getattr(args, a) == None:
                setattr(args, a, cfg.get(a, defaults.get(a)))

        if args.jobs is ...:
            args.jobs = os.cpu_count()
        elif args.jobs <= 0:
            parser.error("--jobs argument must be greater than zero!")

        envWhiteList = recipes.envWhiteList()
        envWhiteList |= set(args.white_list)

        if develop:
            nameFormatter = recipes.getHook('developNameFormatter')
            developPersister = DevelopDirOracle(nameFormatter, recipes.getHook('developNamePersister'))
            nameFormatter = developPersister.getFormatter()
        else:
            nameFormatter = recipes.getHook('releaseNameFormatter')
            nameFormatter = LocalBuilder.releaseNamePersister(nameFormatter)
        nameFormatter = LocalBuilder.makeRunnable(nameFormatter)
        packages = recipes.generatePackages(nameFormatter, defines, args.sandbox)
        if develop: developPersister.prime(packages)

        verbosity = cfg.get('verbosity', 0) + args.verbose - args.quiet
        setVerbosity(verbosity)
        builder = LocalBuilder(recipes, verbosity, args.force,
                               args.no_deps, True if args.build_mode == 'build-only' else False,
                               args.preserve_env, envWhiteList, bobRoot, args.clean,
                               args.no_logfiles)

        builder.setArchiveHandler(getArchiver(recipes))
        builder.setUploadMode(args.upload)
        builder.setDownloadMode(args.download)
        builder.setCleanCheckout(args.clean_checkout)
        builder.setAlwaysCheckout(args.always_checkout + cfg.get('always_checkout', []))
        builder.setLinkDependencies(args.link_deps)
        builder.setJobs(args.jobs)
        builder.setKeepGoing(args.keep_going)
        builder.setAudit(args.audit)
        if args.resume: builder.loadBuildState()

        backlog = []
        providedBacklog = []
        results = []
        for p in args.packages:
            for package in packages.queryPackagePath(p):
                packageStep = package.getPackageStep()
                backlog.append(packageStep)
                # automatically include provided deps when exporting
                build_provided = (args.destination and args.build_provided == None) or args.build_provided
                if build_provided: providedBacklog.extend(packageStep._getProvidedDeps())

        success = runHook(recipes, 'preBuildHook',
            ["/".join(p.getPackage().getStack()) for p in backlog])
        if not success:
            raise BuildError("preBuildHook failed!",
                help="A preBuildHook is set but it returned with a non-zero status.")
        success = False
        if args.jobs > 1:
            setTui(args.jobs)
            builder.enableBufferedIO()
        try:
            builder.cook(backlog, True if args.build_mode == 'checkout-only' else False)
            for p in backlog:
                resultPath = p.getWorkspacePath()
                if resultPath not in results:
                    results.append(resultPath)
            builder.cook(providedBacklog, True if args.build_mode == 'checkout-only' else False, 1)
            for p in providedBacklog:
                resultPath = p.getWorkspacePath()
                if resultPath not in results:
                    results.append(resultPath)
            success = True
        finally:
            if args.jobs > 1: setTui(1)
            builder.saveBuildState()
            runHook(recipes, 'postBuildHook', ["success" if success else "fail"] + results)

    # tell the user
    if results:
        if len(results) == 1:
            print("Build result is in", results[0])
        else:
            print("Build results are in:\n  ", "\n   ".join(results))

        endTime = time.time()
        stats = builder.getStatistic()
        activeOverrides = len(stats.getActiveOverrides())
        print("Duration: " + str(datetime.timedelta(seconds=(endTime - startTime))) + ", "
                + str(stats.checkouts)
                    + " checkout" + ("s" if (stats.checkouts != 1) else "")
                    + " (" + str(activeOverrides) + (" overrides" if (activeOverrides != 1) else " override") + " active), "
                + str(stats.packagesBuilt)
                    + " package" + ("s" if (stats.packagesBuilt != 1) else "") + " built, "
                + str(stats.packagesDownloaded) + " downloaded.")

        # copy build result if requested
        ok = True
        if args.destination:
            for result in results:
                ok = copyTree(result, args.destination) and ok
        if not ok:
            raise BuildError("Could not copy everything to destination. Your aggregated result is probably incomplete.")
    else:
        print("Your query matched no packages. Naptime!")

def doBuild(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob build", description='Build packages in release mode.')
    commonBuildDevelop(parser, argv, bobRoot, False)

def doDevelop(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob dev", description='Build packages in development mode.')
    commonBuildDevelop(parser, argv, bobRoot, True)




# Bob build tool
# Copyright (C) 2017 Ralf Hubert
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .. import BOB_VERSION
from ..input import RecipeSet
from ..tty import colorize
from ..utils import processDefines
import argparse
import asyncio
import json
import re
import os
import collections
from shutil import copyfile

def run(coro):
    return asyncio.get_event_loop().run_until_complete(coro)

def findPackages(package, excludes, highlights, done, maxdepth, donePackages, level = 0):
    def isHighLigh(package):
        for h in highlights:
            if (h.search(package.getName())):
                return True
        return False

    # used for coloring packages in the graph
    # id: 1 - root packages / packages queried for
    #     2 - packages matching the 'highlight' argument
    #     3 - all other packages
    def getColorId(package, level):
        if package.getRecipe().isRoot():
            return 1
        elif level == 0:
            return 1
        elif isHighLigh(package):
            return 2
        return 3

    if package._getId() in donePackages:
       return
    donePackages.add(package._getId())

    for e in excludes:
        if (e.search(package.getName())):
            return

    if package.getPackageStep().isValid():
        if package._getId() not in done:
            done.add(package._getId())
            yield ('node', package, getColorId(package, level))

    if (maxdepth != -1) and (level >= maxdepth):
        return

    for d in package.getDirectDepSteps():
        for e in excludes:
            if (e.search(d.getPackage().getName())):
                break
        else:
           yield ('link', d.getPackage(), package)
        yield from findPackages(d.getPackage(), excludes, highlights, done, maxdepth, donePackages, level+1)

def makeD3Graph(packages, p, filename, options, excludes, highlights, maxdepth):
    def getHover(package, options):
        hover = collections.OrderedDict(package.getMetaEnv())
        if options.get('d3.showScm', False) and package.getCheckoutStep().isValid():
            scmidx = 0
            for s in package.getCheckoutStep().getScmList():
                p = s.getProperties()
                hover['scm'+str(scmidx)] = p.get('scm')
                hover['url'+str(scmidx)] = p.get('url')
                scmidx = scmidx + 1
        return hover

    nodes = []
    links = []
    done = set()
    donePackages = set()
    for package in packages.queryPackagePath(p):
        for (t, a, b) in findPackages(package, excludes, highlights, done, maxdepth, donePackages):
            if t == 'node':
                # do we already have this package? if so add a suffix
                maxVariant = -1
                for i, n in enumerate(nodes):
                    split = n['label'].split('#')
                    if len(split) == 2 and split[0] == a.getName():
                        if int(split[1]) > maxVariant: maxVariant = int(split[1])
                    elif split[0] == a.getName():
                        n['label'] = split[0]+'#0'
                        nodes[i] = n
                nodes.append({"id" : a._getId(),
                    "label" : a.getName() + (('#'+str(maxVariant+1)) if maxVariant != -1 else ''),
                    "colorid" : b,
                    "hover" : json.dumps(getHover(a, options))})
            elif t == 'link':
                link = {"target": a._getId(), "source" : b._getId()}
                if link not in links:
                    links.append(link)

    localLib = options.get('d3.localLib', None)
    if localLib:
        try:
            copyfile(localLib, os.path.join(os.path.dirname(filename),os.path.basename(localLib)))
        except FileNotFoundError as e:
            print(colorize("WARNING: local D3 lib not found - using upstream link ({})".format(e), "33"))
            localLib = None

    with open(filename + ".html", "w") as f:
        f.write(
"""<!DOCTYPE html>
<head>
<title>Bob graph: """ + p + """ </title>
<style>
 .info { font-family: arial; font-size: 10pt;}
</style
</head>
<meta charset="utf-8">
<svg width="0" height="0"></svg>
Highlight packages depending on the selected: <input type="checkbox" id="highlightAll">
<script src='""")
        f.write(os.path.basename(localLib) if localLib else "https://d3js.org/d3.v4.min.js")
        f.write(
"""'></script>
<script>
var nodes =
""")
        json.dump(nodes, f)
        f.write("\nvar links =")
        json.dump(links, f)
        f.write("""\n

function getDependencies(node) {
  return links.reduce(function (neighbors, link) {
      if (link.source.id === node.id) {
        neighbors.push(link.target.id)
      }
      return neighbors
    },
    [node.id]
  )
}

function isDependency(node, link) {
  if ( document.getElementById("highlightAll").checked == true) {
    return (link.source.id === node.id || link.target.id === node.id )
  } else {
    return link.source.id === node.id
  }
}

function getNodeColor(node, neighbors) {
  if (Array.isArray(neighbors) && neighbors.indexOf(node.id) > -1) {
   switch(node.colorid) {
    case 1:
        return 'blue';
    case 2:
        return 'pink';
    default:
        return 'green';
   }
  }

  switch(node.colorid) {
    case 1:
        return 'red';
    case 2:
        return 'orange';
    default:
        return 'gray';
  }
}

function getArrow(node, link) {
   if (( document.getElementById("highlightAll").checked == true) && ( link.target.id === node.id )) {
       return 'url(#arrowhead_dark_magenta)';
   } else if (link.source.id === node.id ) {
       return 'url(#arrowhead_green)';
   } else {
       return 'url(#arrowhead_gray)';
   }
}

function getLinkColor(node, link) {
   if (( document.getElementById("highlightAll").checked == true) && ( link.target.id === node.id )) {
       return '#8B008B';
   } else if (link.source.id === node.id ) {
       return 'green';
   } else {
       return '#E5E5E5'
   }
}


var width = window.innerWidth
var height = window.innerHeight

var zoom = d3.zoom()
    .scaleExtent([0.01, 10])
    .on("zoom", zoomed);

var svg = d3.select('svg')
var container = svg.append("g")

svg.call(zoom)

window.addEventListener('resize', function(event){
    // resize the svg if the window size changes
    var width = window.innerWidth
    var height = window.innerHeight > 200 ? window.innerHeight - 80 : window.innerHeight
    svg.attr('width', width).attr('height', height)
});

// simulation setup with all forces
var linkForce = d3
  .forceLink()
  .id(function (link) { return link.id }).distance(80).strength(0.5)

var attractForce = d3.forceManyBody().strength(1).distanceMax(800)
                     .distanceMin(80);

var collisionForce = d3.forceCollide(80).strength(0.5).iterations(20);

var simulation = d3
  .forceSimulation()
  .nodes(nodes)
  .force('collide', collisionForce)
  .force('link', linkForce)
  .force("center", d3.forceCenter(width / 2, height / 2))

function updateNodes() {
  var neighbors = getDependencies(lastSelected)

  // we modify the styles to highlight selected nodes
  nodeElements.attr('fill', function (node) { return getNodeColor(node, neighbors) })
  linkElements.attr('stroke', function (link) { return getLinkColor(lastSelected, link) })
  linkElements.attr('marker-end', function (link) { return getArrow(lastSelected, link) })
}

function selectNode(selectedNode) {
  lastSelected = selectedNode
  updateNodes();
}

// Create Event Handlers for mouse
function handleMouseOver(d, i) {
  var meta = JSON.parse(d.hover);
  var entries = Object.keys(meta).length

  if (entries != 0) {
    // find the longest entry
    var max = 0
    var lineHeight = 12
    for (e in meta) {
        if (e.length + meta[e].length > max) {
            max = e.length + meta[e].length
        }
    }

    // draw a box
    container.append("rect")
        .attr('id', "tb" + d.id)
        .attr("x", d.x + this.getComputedTextLength())
        .attr("y", d.y - 13)
        .attr("width",  (max+2) * 6 + 20 )
        .attr("height", entries * lineHeight + 10 )
        .attr("rx", 5)
        .attr("ry", 5)
        .attr("style", "fill:white;stroke:black;stroke-width:1")

    var line = 0
    for (e in meta) {
        // add the text
        container
            .append("text")
            .attr('id', "t" + d.id)  // Create an id for text so we can select it later for removing on mouseout
            .attr('x', d.x + this.getComputedTextLength() + 10)
            .attr('y', d.y + line*lineHeight)
            .text(e +': ' + meta[e])  // Value of the text
                .attr("font-size", 10)
                .attr("font-family", "Courier New")
        line += 1
    }
  }
}

function handleMouseOut(d, i) {
    d3.selectAll("#t" + d.id).remove();  // Remove text location
    d3.select("#tb" + d.id).remove();  // Remove textbox location
}

function zoomed() {
  container.attr("transform", d3.event.transform);
}

// Arrow marker for end-of-line arrow
svg.append('defs').append('marker')
    .attr('id', 'arrowhead_gray')
    .attr('refX', 100)
    .attr('refY',   5)
    .attr('markerWidth', 15)
    .attr('markerHeight',15)
    .attr('orient', 'auto')
    .attr('fill', '#E5E5E5')
    .append('path')
    .attr('d', 'M 0,0 l 10,5 l -10,5 Z');

svg.append('defs').append('marker')
    .attr('id', 'arrowhead_green')
    .attr('refX', 100)
    .attr('refY',  5)
    .attr('markerWidth', 15)
    .attr('markerHeight',15 )
    .attr('orient', 'auto')
    .attr('fill', 'green')
    .append('path')
    .attr('d', 'M 0,0 l 10,5 l -10,5 Z');

svg.append('defs').append('marker')
    .attr('id', 'arrowhead_dark_magenta')
    .attr('refX', 100)
    .attr('refY',  5)
    .attr('markerWidth', 15)
    .attr('markerHeight',15 )
    .attr('orient', 'auto')
    .attr('fill', '#8B008B')
    .append('path')
    .attr('d', 'M 0,0 l 10,5 l -10,5 Z');

var linkElements = container.attr("class", "links")
  .selectAll("line")
  .data(links)
  .enter().append("line")
    .attr("stroke-width", 1)
    .attr("stroke", "rgba(50, 50, 50, 0.2)")
    .attr("marker-end", 'url(#arrowhead_gray)');

var nodeElements = container.attr("class", "nodes")
  .selectAll("rect")
  .data(nodes)
  .enter().append("rect")
  .attr("width", function (node) {
          if (node.label.length >= 20) {
             return 130
          } else {
             return node.label.length * 6 + 10
         }})
  .attr("height", function (node) {
          if (node.label.length >= 20) {
             return Math.floor((node.label.length+19) / 20) * 14
          } else {
             return 14
          }})
  .attr("rx", 2)
  .attr("ry", 2)
    .attr("fill", getNodeColor)
  .on('click', selectNode)""")

        if options.get('d3.dragNodes', False):
            f.write("""
  .call(d3.drag()
     .on("start", dragstarted)
     .on("drag", dragged)
     .on("end", dragended));

function dragstarted(d) {
  if (!d3.event.active) simulation.alphaTarget(0.3).restart();
  d.fx = d.x;
  d.fy = d.y;
}

function dragged(d) {
  d.fx = d3.event.x;
  d.fy = d3.event.y;
}

function dragended(d) {
  if (!d3.event.active) simulation.alphaTarget(0);
  d.fx = null;
  d.fy = null;
}""")
        f.write("""\n
var textElements = container.attr("class", "texts")
  .selectAll("text")
  .data(nodes)
  .enter().append("text")
    .text(function (node) { return  node.label })
    .attr("font-size", 10)
    .attr("font-family", "Courier New")
    .attr("dx", 5)
    .attr("dy", 10)
    .on('click', selectNode)
    .on("mouseover", handleMouseOver)
    .on("mouseout", handleMouseOut)
    .call(wrap)""")
        if options.get('d3.dragNodes', False):
            f.write("""
    .call(d3.drag()
     .on("start", dragstarted)
     .on("drag", dragged)
     .on("end", dragended));
""")
        f.write("""\n
function wrap(text) {
   text.each(function() {
        var text = d3.select(this),
        words = text.text().match(/.{1,20}/g).reverse(),
        word,
        line = [],
        lineNumber = 0,
        lineHeight = 1.2,
        x = text.attr("x"),
        y = text.attr("y"),
        dy = text.attr("dy") ? text.attr("dy") : 0;
        tspan = text.text(null).append("tspan");
        while (word = words.pop()) {
            line.push(word);
            tspan.text(line.join(" "));
            if (tspan.node().getComputedTextLength() > 125) {
                line.pop();
                tspan.text(line.join(" "));
                line = [word];
                tspan = text.append("tspan").attr("dx", -120).attr("dy", lineHeight + dy + "em").text(word);
            }
        }
    });
}

var busy = document.createElement('div');
busy.setAttribute("id","busy")
busy.innerHTML = "<p style='font-size:20pt'> Please wait...</p>"
document.body.appendChild(busy);
d3.select("#highlightAll").on("change",updateNodes);
simulation.nodes(nodes).on('end', ticked)
function ticked() {
  busy = document.getElementById("busy")
  if (busy) {
    simulation.nodes(nodes).on('tick', ticked)
    busy.remove();
    // do not make the svg as heigh as the window to see the footer
    svg.attr('width', width).attr('height', height > 200 ? height - 80 : height)
  }
  nodeElements
    .attr('x', function (node) { return node.x })
    .attr('y', function (node) { return node.y })
  textElements
    .attr('x', function (node) { return node.x })
    .attr('y', function (node) { return node.y })
  linkElements
    .attr('x1', function (link) {
        if (link.source.label.length > 21) {
            return link.source.x + 65
        } else {
            return link.source.x + link.source.label.length * 3
        }
    })
    .attr('y1', function (link) {
            return link.source.y + ((link.source.label.length + 20) / 21-1) * 12
        })
    .attr('x2', function (link) {
        if (link.target.label.length > 21) {
            return link.target.x + 65
        } else {
            return link.target.x + link.target.label.length * 3 }
        })
    .attr('y2', function (link) {
            return link.target.y + ((link.target.label.length + 20)/21-1) * 12
        })
}

simulation.force("link").links(links)
</script>
<div class='info' style="width: 100%; text-align: center;">
  <div id='innerLeft' style="float: left"> Recipes: """ + run(RecipeSet().getScmStatus()) + """</div>
  <div id='innerRight' style="float: right">Bob version: """ + BOB_VERSION +"""</div>
  <div id='innerMiddle' style="display: inline-block">Generated using <a href="https://www.d3js.org">D3JS</a></div>
</div>
""")

def makeDotGraph(packages, p, filename, excludes, highlights, maxdepth):
    def getDotColor(c):
        if c == 1:
            return 'red';
        elif c == 2:
            return 'orange';
        else:
            return 'white';

    done = set()
    donePackages = set()
    links = []
    with open(filename + ".dot", "w") as f:
        f.write("digraph \"" + p + "\" {\n")
        f.write(" {\n")
        for package in packages.queryPackagePath(p):
            for (t, a, b) in findPackages(package, excludes, highlights, done, maxdepth, donePackages):
                if t == 'link':
                    if (a.getName(),b.getName()) not in links:
                        links.append((a.getName(), b.getName()))
                elif t == 'node':
                    f.write('   "' + a.getName() + '" [fillcolor=' + getDotColor(b) + ',style=filled]\n')
        f.write(" }\n")
        for (s,d) in links:
            f.write('"{}" -> "{}";\n'.format(d, s))
        f.write("}\n")

def doGraph(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob graph", description='Generate dependency graph')
    parser.add_argument('packages', metavar='PACKAGE', type=str, nargs='+',
        help="Graph entry (sub-)package")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=False,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")
    parser.add_argument('--destination', metavar="DEST",
        help="Destination of graph files.")
    parser.add_argument('-e', '--exclude', default=[], action='append', dest="excludes",
        help="Do not show packages matching this regex. (And all it's deps)")
    parser.add_argument('-f', '--filename', default=None,
        help="Name of Outputfile.")
    parser.add_argument('-H', '--highlight', default=[], action='append', dest="highlights",
        help="Highlight packages matching this regex.")
    parser.add_argument('-n', '--max-depth', type=int, default=None, help="Max depth.")
    parser.add_argument('-t', '--type', choices=['d3', 'dot'], default=None,
        help="Graph type. d3 (default) or dot")
    parser.add_argument("-o", default=[], action='append', dest='options',
        help="Set extended options")
    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    recipes = RecipeSet()
    recipes.setConfigFiles(args.configFile)
    recipes.parse()
    packages = recipes.generatePackages(lambda s,m: "unused", defines, args.sandbox)

    cfg = recipes.getCommandConfig().get('graph', {})
    defaults = {
        'type' : 'd3',
        'max_depth' : -1,
    }

    options = cfg.get('options', {})
    for i in args.options:
        (opt, sep, val) = i.partition("=")
        if sep != "=":
            parser.error("Malformed plugin option: "+i)
        if val != "":
            options[opt] = val

    for a in vars(args):
       if a=='options': continue
       if getattr(args, a) == None:
            setattr(args, a, cfg.get(a, defaults.get(a)))

    excludes = []
    if args.excludes:
        for e in args.excludes:
            excludes.append(re.compile(e))

    highlights = []
    if args.highlights:
        for e in args.highlights:
            highlights.append(re.compile(e))

    destination = os.path.join(os.getcwd(), 'graph') if not args.destination else args.destination
    if not os.path.exists(destination):
        os.makedirs(destination)

    for p in args.packages:
        # convert the package-query into a valid file name
        filename = args.filename if args.filename else "".join([c for c in p if c.isalpha() or c.isdigit() or c==' ']).rstrip()

        print(">>", colorize(p, "32;1"))
        print(colorize("   GRAPH   {} ({})".format(p, args.type), "32"))

        if args.type == 'd3':
            makeD3Graph(packages, p, os.path.join(destination, filename), options, excludes, highlights, args.max_depth)
        elif args.type == 'dot':
            makeDotGraph(packages, p, os.path.join(destination, filename), excludes, highlights, args.max_depth)



# -*- coding: utf-8 -*-

# Bob build tool
# Copyright (C) 2016  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..input import RecipeSet
from ..errors import BuildError
from ..utils import processDefines
import argparse
import codecs
import sys

try:
    # test if stdout can handle box drawing characters
    codecs.encode("└├│─", sys.stdout.encoding)
    LS_SEP_1 = u"└── "
    LS_SEP_2 = u"├── "
    LS_SEP_3 = u"    "
    LS_SEP_4 = u"│   "
except UnicodeEncodeError:
    # fall back to ASCII
    LS_SEP_1 = "\\-- "
    LS_SEP_2 = "|-- "
    LS_SEP_3 = "    "
    LS_SEP_4 = "|   "


class PackagePrinter:
    def __init__(self, showAll, showOrigin, recurse, unsorted):
        self.showAll = showAll
        self.showOrigin = showOrigin
        self.recurse = recurse
        if unsorted:
            self.sort = lambda x: x
        else:
            self.sort = sorted

    def __getChilds(self, package):
        return [
            (name, child.node, " ({})".format(child.origin) if (self.showOrigin and child.origin) else "")
            for (name, child) in self.sort(package.items())
            if (self.showAll or child.direct)
        ]

    def showTree(self, package, prefix=""):
        i = 0
        packages = self.__getChilds(package)
        for (n, p, o) in packages:
            last = (i >= len(packages)-1)
            print("{}{}{}{}".format(prefix, LS_SEP_1 if last else LS_SEP_2, n, o))
            self.showTree(p, prefix + (LS_SEP_3 if last else LS_SEP_4))
            i += 1

    def showPrefixed(self, package, showAliases, stack=[], level=0):
        for p in showAliases: print(p)
        for (n, p, o) in self.__getChilds(package):
            newStack = stack[:]
            newStack.append(n)
            print("{}{}".format("/".join(newStack), o))
            if self.recurse:
                self.showPrefixed(p, [], newStack, level+1)


def doLS(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob ls", description='List packages.')
    parser.add_argument('package', type=str, nargs='?', default="",
                        help="Sub-package to start listing from")
    parser.add_argument('-a', '--all', default=False, action='store_true',
                        help="Show indirect dependencies too")
    parser.add_argument('-o', '--origin', default=False, action='store_true',
                        help="Show origin of indirect dependencies")
    parser.add_argument('-r', '--recursive', default=False, action='store_true',
                        help="Recursively display dependencies")
    parser.add_argument('-u', '--unsorted', default=False, action='store_true',
                        help="Show packages in recipe order (unsorted)")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-p', '--prefixed', default=False, action='store_true',
                       help="Prints the full path prefix for each package")
    group.add_argument('-d', '--direct', default=False, action='store_true',
                       help="List packages themselves, not their contents")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=False,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")
    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    recipes = RecipeSet()
    recipes.setConfigFiles(args.configFile)
    recipes.parse()

    packages = recipes.generatePackages(lambda s,m: "unused", defines, args.sandbox)
    showAliases = packages.getAliases() if args.package == "" else []

    printer = PackagePrinter(args.all, args.origin, args.recursive, args.unsorted)
    for (stack, root) in packages.queryTreePath(args.package):
        if args.prefixed:
            printer.showPrefixed(root, showAliases, stack)
        elif args.direct:
            print("/".join(stack) if stack else "/")
        elif args.recursive:
            print("/".join(stack) if stack else "/")
            printer.showTree(root)
        else:
            printer.showPrefixed(root, showAliases)

class Default(dict):
    def __init__(self, default, *args, **kwargs):
        self.__default = default
        super().__init__(*args, **kwargs)

    def __missing__(self, key):
        return self.__default

def doQueryMeta(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob query-meta",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""Query meta information of packages.""")
    parser.add_argument('packages', nargs='+', help="(Sub-)packages to query")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    parser.add_argument('-r', '--recursive', default=False, action='store_true',
                        help="Recursively display dependencies")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=False,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")
    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    recipes = RecipeSet()
    recipes.setConfigFiles(args.configFile)
    recipes.parse()
    packages = recipes.generatePackages(lambda s,m: "unused", defines, args.sandbox)

    def showPackage(package, recurse, done):
        # Show each package only once. Meta variables are fixed and not variant
        # dependent.
        key = package.getName()
        if key not in done:
            for (var, val) in package.getMetaEnv().items():
                print(package.getName() + " " + var + "=" + val)
            done.add(key)

        # recurse package tree if requested
        if recurse:
            for ps in package.getDirectDepSteps():
                showPackage(ps.getPackage(), recurse, done)

    done = set()
    for p in args.packages:
        for package in packages.queryPackagePath(p):
            showPackage(package, args.recursive, done)

def doQuerySCM(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob query-scm",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Query SCM configuration of packages.

By default this command will print one line for each SCM in the given package.
The output format may be overridded by '-f'. By default the following formats
are used:

 * git="git {package} {dir} {url} {branch}"
 * svn="svn {package} {dir} {url} {revision}"
 * cvs="cvs {package} {dir} {cvsroot} {module}"
 * url="url {package} {dir}/{fileName} {url}"
""")
    parser.add_argument('packages', nargs='+', help="(Sub-)packages to query")

    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    parser.add_argument('-f', default=[], action='append', dest="formats",
        help="Output format for scm (syntax: scm=format). Can be specified multiple times.")
    parser.add_argument('--default', default="", help='Default for missing attributes (default: "")')
    parser.add_argument('-r', '--recursive', default=False, action='store_true',
                        help="Recursively display dependencies")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=False,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")

    formats = {
        'git' : "git {package} {dir} {url} {branch}",
        'svn' : "svn {package} {dir} {url} {revision}",
        'cvs' : "cvs {package} {dir} {cvsroot} {module}",
        'url' : "url {package} {dir}/{fileName} {url}",
    }

    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    recipes = RecipeSet()
    recipes.setConfigFiles(args.configFile)
    recipes.parse()
    packages = recipes.generatePackages(lambda s,m: "unused", defines, args.sandbox)

    # update formats
    for fmt in args.formats:
        f = fmt.split("=")
        if len(f) != 2: parser.error("Malformed format: "+fmt)
        formats[f[0]] = f[1]

    def showPackage(package, recurse, done, donePackages):
        if package._getId() in donePackages:
            return
        donePackages.add(package._getId())

        # show recipes only once for each checkout variant
        key = (package.getRecipe().getName(), package.getCheckoutStep().getVariantId())
        if key not in done:
            for scm in package.getCheckoutStep().getScmList():
                p = { k:v for (k,v) in scm.getProperties().items() if v is not None }
                p['package'] = "/".join(package.getStack())
                fmt = formats.get(p['scm'], "{scm} {dir}")
                print(fmt.format_map(Default(args.default, p)))
            done.add(key)

        # recurse package tree if requested
        if recurse:
            for ps in package.getDirectDepSteps():
                showPackage(ps.getPackage(), recurse, done, donePackages)

    done = set()
    donePackages = set()
    for p in args.packages:
        for package in packages.queryPackagePath(p):
            showPackage(package, args.recursive, done, donePackages)

def doQueryRecipe(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob query-recipe",
        description="Query recipe and class files of package.")
    parser.add_argument('package', help="(Sub-)package to query")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=False,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")

    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    recipes = RecipeSet()
    recipes.setConfigFiles(args.configFile)
    recipes.parse()
    package = recipes.generatePackages(lambda s,m: "unused", defines, args.sandbox).walkPackagePath(args.package)

    for fn in package.getRecipe().getSources():
        print(fn)



# Bob build tool
# Copyright (C) 2019  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..errors import BuildError
from ..utils import removePath, isWindows, summonMagic
import argparse
import os
import re
import sys

__all__ = ('parseArgumentLine', 'CommonIDEGenerator')

if isWindows():
    INVALID_CHAR_TRANS = str.maketrans(':*?<>"|', '_______')
else:
    INVALID_CHAR_TRANS = str.maketrans('', '')
EXCLUDE_DIRS = frozenset(['.git', '.svn', 'CVS'])
SOURCE_FILES = re.compile(r"\.[c](pp)?$", re.IGNORECASE)
HEADER_FILES = re.compile(r"\.[h](pp)?$", re.IGNORECASE)
RESOURCE_FILES = re.compile(r"\.cmake$\|^CMakeLists.txt$", re.IGNORECASE)

def filterDirs(directories):
    i = 0
    while i < len(directories):
        if directories[i] in EXCLUDE_DIRS:
            del directories[i]
        else:
            i += 1

def parseArgumentLine(line):
    lines = []
    line.lstrip()
    if line.startswith('@'):
        filename = line.lstrip('@')
        try:
            lines = [line.rstrip('\n') for line in open(filename)]
        except IOError:
            raise BuildError("Input file (" + filename + ") could not be read.",
                             help="Specify the argument without leading @, if it is not a file path.")
    else:
        lines.append(line)
    return lines

class BaseScanner:
    def __init__(self, isRoot=False, stack="", additionalFiles=[]):
        self.isRoot = isRoot
        self.stack = stack
        self.workspacePath = ''
        self.__additionalFiles = additionalFiles
        self.__headers = set()
        self.__sources = set()
        self.__resources = set()
        self.__incPaths = set()
        self.__dependencies = set()
        self.__runTargets = []

    def _addFile(self, root, fileName):
        added = True

        if SOURCE_FILES.search(fileName):
            self.__sources.add(os.path.join(root, fileName))
        elif HEADER_FILES.search(fileName):
            self.__headers.add(os.path.join(root, fileName))
        elif RESOURCE_FILES.search(fileName):
            self.__resources.add(os.path.join(root, fileName))
        elif self.__additionalFiles and self.__additionalFiles.search(fileName):
            self.__resources.add(os.path.join(root, fileName))
        else:
            added = False

        return added

    def _addIncDirectory(self, directory):
        self.__incPaths.add(directory)

    def _addDepedency(self, dependency):
        self.__dependencies.add(dependency)

    def _addRunTarget(self, target):
        self.__runTargets.append(target)

    def scan(self, workspacePath):
        self.workspacePath = workspacePath
        return False

    @property
    def headers(self):
        return sorted(self.__headers)

    @property
    def sources(self):
        return sorted(self.__sources)

    @property
    def resources(self):
        return sorted(self.__resources)

    @property
    def incPaths(self):
        return sorted(self.__incPaths)

    @property
    def dependencies(self):
        return sorted(self.__dependencies)

    @property
    def runTargets(self):
        return self.__runTargets


class GenericScanner(BaseScanner):
    def __init__(self, isRoot, stack, additionalFiles):
        super().__init__(isRoot, stack, additionalFiles)

    def scan(self, workspacePath):
        super().scan(workspacePath)
        ret = False

        for root, directories, filenames in os.walk(workspacePath):
            # remove scm directories, os.walk will only descend into directiries
            # that stay in the list
            filterDirs(directories)
            hasInclude = False
            for filename in filenames:
                ret = self._addFile(root, filename) or ret
                hasInclude = hasInclude or HEADER_FILES.search(filename)
            if hasInclude:
                oldPath = ""
                # need to recursively add all directories from the include up to cwd to get includes like
                # <a/b/c.h> resolved even if there is no include in 'a'
                relativePath = root[len(workspacePath):]
                for path in relativePath.split(os.path.sep):
                    includePath = os.path.join(workspacePath, oldPath, path)
                    self._addIncDirectory(includePath)
                    oldPath = os.path.join(oldPath, path)

        return ret

SCANNERS = [
    GenericScanner  # must be last as "catch all" last resort
]

class CommonIDEGenerator:

    def __init__(self, prog, description):
        parser = argparse.ArgumentParser(prog="bob project " + prog,
                                         description=description)
        self.parser = parser

        parser.add_argument('--name', metavar="NAME",
            help="Name of project (default: package name).")
        parser.add_argument('--destination', metavar="DEST",
            help="Destination of project files (default: projects/NAME).")
        parser.add_argument('--overwrite', action='store_true',
            help="Remove destination folder before generating.")
        parser.add_argument('-u', '--update', default=False, action='store_true',
            help="Update project files")

        parser.add_argument('--exclude', default=[], action='append', dest="excludes",
                help="Package filter. A regex for excluding packages in QTCreator.")
        parser.add_argument('--include', default=[], action='append', dest="include",
                help="Include package filter. A regex for including only the specified packages in QTCreator.")

        parser.add_argument('-I', dest="additional_includes", default=[], action='append',
            help="Additional include directories. (added recursive starting from this directory).")
        parser.add_argument('-S', dest="start_includes", default=[], action='append',
            help="Additional include directories, will be placed at the beginning of the include list.")
        parser.add_argument('-f', '--filter', metavar="Filter",
            help="File filter. A regex for matching additional files.")

    def __walk(self, package, excludes, additionalFiles):
        class CheckoutInfo:
            def __init__(self, scan, packageVid):
                self.scan = scan
                self.packages = set([packageVid])
                self.name = None

        class PackageInfo:
            def __init__(self, recipeName, packageName):
                self.recipeName = recipeName
                self.packageName = packageName
                self.checkout = False
                self.dependencies = set()

        checkouts = { False : CheckoutInfo(None, None) }
        packages = {}

        def collect(package, rootPackage=True):
            # only once per package
            packageVid = package.getPackageStep().getVariantId()
            if packageVid in packages: return
            packages[packageVid] = packageInfo = PackageInfo(package.getRecipe().getName(), package.getName())

            # only scan if checkout step is valid and was not already scanned
            checkoutPath = package.getCheckoutStep().isValid() and \
                package.getCheckoutStep().getWorkspacePath()
            if checkoutPath not in checkouts:
                # take result from first scanner that succeeds
                for scanner in SCANNERS:
                    scan = scanner(rootPackage, "/".join(package.getStack()), additionalFiles)
                    if scan.scan(checkoutPath):
                        break
                else:
                    scan = BaseScanner()
                info = checkouts[checkoutPath] = CheckoutInfo(scan, packageVid)
            elif rootPackage:
                # make sure that root package has at least an empty checkout to be always visible
                checkoutPath = "<root-sentinel>"
                info = checkouts[checkoutPath] = CheckoutInfo(BaseScanner(True, "/".join(package.getStack())), packageVid)
            else:
                info = checkouts[checkoutPath]
                info.packages.add(packageVid)

            # find all executables
            if package.getPackageStep().isValid():
                packageDir = package.getPackageStep().getWorkspacePath()
                if sys.platform in ["msys", "win32"]:
                    for root, directory, filenames in os.walk(packageDir):
                        for filename in filenames:
                            target = os.path.join(root, filename)
                            if target.lower().endswith(".exe"):
                                info.scan._addRunTarget(target)
                else:
                    magic = summonMagic()
                    for root, directory, filenames in os.walk(packageDir):
                        for filename in filenames:
                            target = os.path.join(root, filename)
                            try:
                                ftype = magic.from_file(target)
                                if 'executable' in ftype and 'x86' in ftype:
                                    info.scan._addRunTarget(target)
                            except OSError:
                                pass

            # descend on used dependencies
            packageInfo.checkout = checkoutPath
            for d in package.getBuildStep().getArguments():
                depName = d.getPackage().getName()
                if any(e.search(depName) for e in excludes):
                    continue
                if d.getPackage() == package:
                    continue
                packageInfo.dependencies.add(d.getVariantId())
                collect(d.getPackage(), False)

        # Collect all packages and their respective checkouts
        collect(package)
        del checkouts[False] # remove sentinel

        # Loop through all checkouts and assign a name. If there is a single
        # package using it then we take the package name. Otherwise we take the
        # recipe name. If multiple checkouts have the same name we add a
        # counting suffix.
        names = set()
        for checkoutPath, info in sorted(checkouts.items()):
            checkoutPackages = [ packages[vid] for vid in sorted(info.packages) ]
            if len(checkoutPackages) > 1:
                # pick the first name if multiple recipes refer to the same checkout
                name = sorted(set(p.recipeName for p in checkoutPackages))[0]
            else:
                name = checkoutPackages[0].packageName

            name = name.translate(INVALID_CHAR_TRANS)
            suffix = ""
            num = 1
            while name+suffix in names:
                suffix = "-{}".format(num)
                num += 1
            name += suffix
            names.add(name)
            info.name = name

        # Walk all packages
        # FIXME: prevent cyclic dependencies
        result = {}
        done = set()
        todo = [ package.getPackageStep().getVariantId() ]
        while todo:
            vid = todo.pop(0)
            if vid in done: continue
            info = packages[vid]
            done.add(vid)

            co = checkouts.get(info.checkout)
            if co:
                result[co.name] = co.scan
                for dep in info.dependencies:
                    depCo = checkouts.get(packages[dep].checkout)
                    if depCo is not None: co.scan._addDepedency(depCo.name)

            todo.extend(d for d in sorted(info.dependencies) if d not in done)

        return result

    def configure(self, package, argv):
        self.rootPackage = package
        self.args  = self.parser.parse_args(argv)

        self.projectName = (self.args.name or package.getName()).translate(INVALID_CHAR_TRANS)
        self.destination = (self.args.destination or
            os.path.join("projects", self.projectName)).translate(INVALID_CHAR_TRANS)

        excludes = []
        try:
            for e in self.args.excludes: excludes.append(re.compile(e))
            for e in self.args.include: excludes.append(re.compile(r"^((?!"+e+").)*$"))
            additionalFiles = re.compile(args.filter) if self.args.filter else None
        except re.error as e:
            raise ParseError("Invalid regular expression '{}': {}".format(e.pattern), e)

        self.packages = self.__walk(package, excludes, additionalFiles)

        # include directories that are statically prepended
        self.prependIncludeDirectories = []
        for i in self.args.start_includes:
            for e in parseArgumentLine(i):
                if os.path.exists(e):
                    self.prependIncludeDirectories.append(e)

        # include directories that are statically appended
        self.appendIncludeDirectories = []
        for i in self.args.additional_includes:
            for e in parseArgumentLine(i):
                if os.path.exists(e):
                    for root, directories, filenames in os.walk(e):
                        filterDirs(directories)
                        self.appendIncludeDirectories.append(os.path.join(e,root))

    def generate(self):
        if self.args.overwrite:
            removePath(self.destination)
        if not os.path.exists(self.destination):
            os.makedirs(self.destination)

    def updateFile(self, name, content, encoding=None, newline=None):
        newName = name+".new"
        oldName = name+".old"
        with open(newName, "w", encoding=encoding, newline=newline) as f:
            f.write(content)
        with open(newName, "rb") as f:
            newContent = f.read()
        try:
            with open(oldName, "rb") as f:
                oldContent = f.read()
        except OSError:
            oldContent = None

        if oldContent != newContent:
            os.replace(newName, name)
            with open(oldName, "wb") as f:
                f.write(newContent)
        else:
            os.remove(newName)



# Bob build tool
# Copyright (C) 2016-2018  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ... import BOB_VERSION
from ...archive import DummyArchive
from ...audit import Audit
from ...errors import BobError, BuildError, MultiBobError
from ...input import RecipeSet
from ...invoker import Invoker, InvocationMode
from ...languages import StepSpec
from ...state import BobState
from ...stringparser import Env
from ...tty import log, stepMessage, stepAction, stepExec, setProgress, ttyReinit, \
    SKIPPED, EXECUTED, INFO, WARNING, DEFAULT, \
    ALWAYS, IMPORTANT, NORMAL, INFO, DEBUG, TRACE
from ...utils import asHexStr, hashDirectory, removePath, emptyDirectory, \
    isWindows, INVALID_CHAR_TRANS, quoteCmdExe, getPlatformTag
from shlex import quote
from textwrap import dedent
import argparse
import asyncio
import concurrent.futures
import datetime
import hashlib
import io
import locale
import os
import re
import shutil
import signal
import stat
import sys
import tempfile

# Output verbosity:
#    <= -2: package name
#    == -1: package name, package steps
#    ==  0: package name, package steps, stderr
#    ==  1: package name, package steps, stderr, stdout
#    ==  2: package name, package steps, stderr, stdout, set -x

async def gatherTasks(tasks):
    if not tasks:
        return []

    await asyncio.wait(tasks)
    return [ t.result() for t in tasks ]

def hashWorkspace(step):
    return hashDirectory(step.getWorkspacePath(),
        os.path.join(step.getWorkspacePath(), "..", "cache.bin"))

def compareDirectoryState(left, right):
    """Compare two directory states while ignoring the SCM specs.

    The SCM specs might change even though the digest stays the same (e.g. the
    URL changes but the commit id stays the same).  This function filters the
    spec to detect real changes.
    """
    left  = { d : v[0] for d, v in left.items()  }
    right = { d : v[0] for d, v in right.items() }
    return left == right

def dissectPackageInputState(oldInputBuildId):
    """Take a package step input hashes and convert them to a common
    representation.

    Excluding the legacy storage, two formats are persisted:

      built: [ BuildId, InputHash1, ...]
      downloaded: BuildId

    Returned tuple:
        (wasDownloaded:bool, inputHashes:list, buildId:bytes)
    """
    if (isinstance(oldInputBuildId, list) and (len(oldInputBuildId) >= 1)):
        oldWasDownloaded = False
        oldInputHashes = oldInputBuildId[1:]
        oldInputBuildId = oldInputBuildId[0]
    elif isinstance(oldInputBuildId, bytes):
        oldWasDownloaded = True
        oldInputHashes = None
    else:
        # created by old Bob version or new workspace
        oldWasDownloaded = False
        oldInputHashes = oldInputBuildId

    return (oldWasDownloaded, oldInputHashes, oldInputBuildId)

def packageInputDownloaded(buildId):
    assert isinstance(buildId, bytes)
    return buildId

def packageInputBuilt(buildId, inputHashes):
    assert isinstance(buildId, bytes)
    return [buildId] + inputHashes


class RestartBuildException(Exception):
    pass

class CancelBuildException(Exception):
    pass

class LocalBuilderStatistic:
    def __init__(self):
        self.__activeOverrides = set()
        self.checkouts = 0
        self.packagesBuilt = 0
        self.packagesDownloaded = 0

    def addOverrides(self, overrides):
        self.__activeOverrides.update(overrides)

    def getActiveOverrides(self):
        return self.__activeOverrides

class LocalBuilder:

    RUN_TEMPLATE_POSIX = """#!/bin/bash
cd {ROOT}
{BOB} _invoke {CLEAN} {SPEC} "$@"
"""

    RUN_TEMPLATE_WINDOWS = """@ECHO OFF
cd {ROOT}
{BOB} _invoke {CLEAN} {SPEC} %*
"""

    @staticmethod
    def releaseNameFormatter(step, props):
        if step.isCheckoutStep():
            base = step.getPackage().getRecipe().getName()
        else:
            base = step.getPackage().getName()
        return os.path.join("work", base.replace('::', os.sep), step.getLabel())

    @staticmethod
    def releaseNamePersister(wrapFmt):

        def fmt(step, props):
            return BobState().getByNameDirectory(
                wrapFmt(step, props),
                asHexStr(step.getVariantId()),
                step.isCheckoutStep())

        return fmt

    @staticmethod
    def releaseNameInterrogator(step, props):
        return BobState().getExistingByNameDirectory(asHexStr(step.getVariantId()))

    @staticmethod
    def developNameFormatter(step, props):
        if step.isCheckoutStep():
            base = step.getPackage().getRecipe().getName()
        else:
            base = step.getPackage().getName()
        return os.path.join("dev", step.getLabel(), base.replace('::', os.sep))

    @staticmethod
    def makeRunnable(wrapFmt):
        baseDir = os.getcwd()

        def fmt(step, mode, props, referrer):
            if mode == 'workspace':
                ret = wrapFmt(step, props)
            else:
                assert mode == 'exec'
                if referrer.getSandbox() is None:
                    ret = os.path.join(baseDir, wrapFmt(step, props))
                else:
                    ret = os.path.join("/bob", asHexStr(step.getVariantId()))
            return os.path.join(ret, "workspace") if ret is not None else None

        return fmt

    def __init__(self, recipes, verbose, force, skipDeps, buildOnly, preserveEnv,
                 envWhiteList, bobRoot, cleanBuild, noLogFile):
        self.__recipes = recipes
        self.__wasRun= {}
        self.__wasSkipped = {}
        self.__wasDownloadTried = {}
        self.__verbose = max(ALWAYS, min(TRACE, verbose))
        self.__noLogFile = noLogFile
        self.__force = force
        self.__skipDeps = skipDeps
        self.__buildOnly = buildOnly
        self.__preserveEnv = preserveEnv
        self.__envWhiteList = set(envWhiteList)
        self.__archive = DummyArchive()
        self.__downloadDepth = 0xffff
        self.__downloadDepthForce = 0xffff
        self.__downloadPackages = None
        self.__bobRoot = bobRoot
        self.__cleanBuild = cleanBuild
        self.__cleanCheckout = False
        self.__srcBuildIds = {}
        self.__buildDistBuildIds = {}
        self.__statistic = LocalBuilderStatistic()
        self.__alwaysCheckout = []
        self.__linkDeps = True
        self.__jobs = 1
        self.__bufferedStdIO = False
        self.__keepGoing = False
        self.__audit = True
        self.__fingerprints = { None : b'', "" : b'' }
        self.__workspaceLocks = {}

    def setArchiveHandler(self, archive):
        self.__archive = archive

    def setDownloadMode(self, mode):
        self.__downloadDepth = 0xffff
        if mode in ('yes', 'forced'):
            self.__archive.wantDownload(True)
            if mode == 'forced':
                self.__downloadDepth = 0
                self.__downloadDepthForce = 0
            elif self.__archive.canDownloadLocal():
                self.__downloadDepth = 0
        elif mode in ('deps', 'forced-deps'):
            self.__archive.wantDownload(True)
            if mode == 'forced-deps':
                self.__downloadDepth = 1
                self.__downloadDepthForce = 1
            elif self.__archive.canDownloadLocal():
                self.__downloadDepth = 1
        elif mode == 'forced-fallback':
            self.__archive.wantDownload(True)
            self.__downloadDepth = 0
            self.__downloadDepthForce = 1
        elif mode.startswith('packages='):
            self.__archive.wantDownload(True)
            try:
                self.__downloadPackages = re.compile(mode[9:])
            except re.error as e:
                raise BuildError("Invalid download regex '{}': {}".format(e.pattern, e))
        else:
            assert mode == 'no'
            self.__archive.wantDownload(False)

    def setUploadMode(self, mode):
        self.__archive.wantUpload(mode)

    def setCleanCheckout(self, clean):
        self.__cleanCheckout = clean

    def setAlwaysCheckout(self, alwaysCheckout):
        try:
            self.__alwaysCheckout = [ re.compile(e) for e in alwaysCheckout ]
        except re.error as e:
            raise BuildError("Invalid --always-checkout regex '{}': {}".format(e.pattern, e))

    def setLinkDependencies(self, linkDeps):
        self.__linkDeps = linkDeps

    def setJobs(self, jobs):
        self.__jobs = max(jobs, 1)

    def enableBufferedIO(self):
        self.__bufferedStdIO = True

    def setKeepGoing(self, keepGoing):
        self.__keepGoing = keepGoing

    def setAudit(self, audit):
        self.__audit = audit

    def saveBuildState(self):
        state = {}
        # Save 'wasRun' as plain dict. Skipped steps are dropped because they
        # were not really executed. Either they are simply skipped again or, if
        # the user changes his mind, they will finally be executed.
        state['wasRun'] = { path : (vid, isCheckoutStep)
            for path, (vid, isCheckoutStep) in self.__wasRun.items()
            if not self.__wasSkipped.get(path, False) }
        # Save all predicted src build-ids. In case of a resume we won't ask
        # the server again for a live-build-id. Regular src build-ids are
        # cached by the usual 'wasRun' and 'resultHash' states.
        state['predictedBuidId'] = { (path, vid) : bid
            for (path, vid), (bid, predicted) in self.__srcBuildIds.items()
            if predicted }
        BobState().setBuildState(state)

    def loadBuildState(self):
        state = BobState().getBuildState()
        self.__wasRun = dict(state.get('wasRun', {}))
        self.__srcBuildIds = { (path, vid) : (bid, True)
            for (path, vid), bid in state.get('predictedBuidId', {}).items() }

    def _wasAlreadyRun(self, step, skippedOk):
        path = step.getWorkspacePath()
        if path in self.__wasRun:
            digest = self.__wasRun[path][0]
            # invalidate invalid cached entries
            if digest != step.getVariantId():
                del self.__wasRun[path]
                return False
            elif (not skippedOk) and self.__wasSkipped.get(path, False):
                return False
            else:
                return True
        else:
            return False

    def _setAlreadyRun(self, step, isCheckoutStep, skipped=False):
        path = step.getWorkspacePath()
        self.__wasRun[path] = (step.getVariantId(), isCheckoutStep)
        self.__wasSkipped[path] = skipped

    def _clearWasRun(self):
        """Clear "was-run" info for build- and package-steps."""
        self.__wasRun = { path : (vid, isCheckoutStep)
            for path, (vid, isCheckoutStep) in self.__wasRun.items()
            if isCheckoutStep }

    def _wasDownloadTried(self, step):
        return self.__wasDownloadTried.get(step.getWorkspacePath(), False)

    def _setDownloadTried(self, step):
        self.__wasDownloadTried[step.getWorkspacePath()] = True

    def _clearDownloadTried(self):
        self.__downloadDisposition = {}

    def _constructDir(self, step, label):
        created = False
        workDir = step.getWorkspacePath()
        if not os.path.isdir(workDir):
            os.makedirs(workDir)
            created = True
        return (workDir, created)

    def __workspaceLock(self, step):
        path = step.getWorkspacePath()
        ret = self.__workspaceLocks.get(path)
        if ret is None:
            self.__workspaceLocks[path] = ret = asyncio.Lock()
        return ret

    async def _generateAudit(self, step, depth, resultHash, executed=True):
        auditPath = os.path.join(step.getWorkspacePath(), "..", "audit.json.gz")
        if os.path.exists(auditPath): removePath(auditPath)
        if not self.__audit:
            return None

        if step.isCheckoutStep():
            buildId = resultHash
        else:
            buildId = await self._getBuildId(step, depth)

        def auditOf(s):
            return os.path.join(s.getWorkspacePath(), "..", "audit.json.gz")

        with stepAction(step, "AUDIT", step.getWorkspacePath(), INFO) as a:
            audit = Audit.create(step.getVariantId(), buildId, resultHash)
            audit.addDefine("bob", BOB_VERSION)
            audit.addDefine("recipe", step.getPackage().getRecipe().getName())
            audit.addDefine("package", "/".join(step.getPackage().getStack()))
            audit.addDefine("step", step.getLabel())
            audit.addDefine("language", step.getPackage().getRecipe().scriptLanguage.index.value)
            for var, val in step.getPackage().getMetaEnv().items():
                audit.addMetaEnv(var, val)
            audit.setRecipesAudit(await step.getPackage().getRecipe().getRecipeSet().getScmAudit())

            # The following things make only sense if we just executed the step
            if executed:
                try:
                    audit.setEnv(os.path.join(step.getWorkspacePath(), "..", "env"))
                    for (name, tool) in sorted(step.getTools().items()):
                        audit.addTool(name, auditOf(tool.getStep()))
                    sandbox = step.getSandbox()
                    if sandbox is not None:
                        audit.setSandbox(auditOf(sandbox.getStep()))
                    for dep in step.getArguments():
                        if dep.isValid(): audit.addArg(auditOf(dep))
                except BobError as e:
                    a.fail(e.slogan, WARNING)
                    return None

            # Always check for SCMs but don't fail if we did not execute the step
            if step.isCheckoutStep():
                for scm in step.getScmList():
                    auditSpec = scm.getAuditSpec()
                    if auditSpec is not None:
                        (typ, dir, extra) = auditSpec
                        try:
                            await audit.addScm(typ, step.getWorkspacePath(), dir, extra)
                        except BobError as e:
                            if executed: raise
                            stepMessage(step, "AUDIT", "WARNING: cannot audit SCM: {} ({})"
                                                .format(e.slogan, dir),
                                           WARNING)

            audit.save(auditPath)
            return auditPath

    def __linkDependencies(self, step):
        """Create symlinks to the dependency workspaces"""

        # this will only work on POSIX
        if isWindows(): return

        if not self.__linkDeps: return

        # always re-create the deps directory
        basePath = os.getcwd()
        depsPath = os.path.join(basePath, step.getWorkspacePath(), "..", "deps")
        removePath(depsPath)
        os.makedirs(depsPath)

        def linkTo(dest, linkName):
            os.symlink(os.path.relpath(os.path.join(basePath, dest, ".."),
                                       os.path.join(linkName, "..")),
                       linkName)

        # there can only be one sandbox
        if step.getSandbox() is not None:
            sandboxPath = os.path.join(depsPath, "sandbox")
            linkTo(step.getSandbox().getStep().getWorkspacePath(), sandboxPath)

        # link tools by name
        tools = step.getTools()
        if tools:
            toolsPath = os.path.join(depsPath, "tools")
            os.makedirs(toolsPath)
            for (n,t) in tools.items():
                linkTo(t.getStep().getWorkspacePath(), os.path.join(toolsPath, n))

        # link dependencies by position and name
        args = step.getArguments()
        if args:
            argsPath = os.path.join(depsPath, "args")
            os.makedirs(argsPath)
            i = 1
            for a in args:
                if a.isValid():
                    linkTo(a.getWorkspacePath(),
                           os.path.join(argsPath,
                                        "{:02}-{}".format(i, a.getPackage().getName())))
                i += 1

    async def _runShell(self, step, scriptName, logger, cleanWorkspace=None):
        workspacePath = step.getWorkspacePath()
        if not os.path.isdir(workspacePath): os.makedirs(workspacePath)
        self.__linkDependencies(step)

        # write spec
        specFile = os.path.join(workspacePath, "..", "step.spec")
        envFile = os.path.join(workspacePath, "..", "env")
        logFile = os.path.join(workspacePath, "..", "log.txt")
        scriptHint = os.path.join(workspacePath, "..", "script")
        spec = StepSpec.fromStep(step, envFile, self.__envWhiteList, logFile,
            scriptHint=scriptHint)
        with open(specFile, "w") as f:
            spec.toFile(f)

        # write invocation wrapper
        if sys.platform == "win32":
            runFile = os.path.join("..", scriptName + ".cmd")
            runFileContent = self.RUN_TEMPLATE_WINDOWS.format(
                    ROOT=quoteCmdExe(os.getcwd()),
                    BOB=quoteCmdExe(self.__bobRoot),
                    SPEC=quoteCmdExe(specFile),
                    CLEAN="-c" if cleanWorkspace else "",
                )
        else:
            runFile = os.path.join("..", scriptName + ".sh")
            runFileContent = self.RUN_TEMPLATE_POSIX.format(
                    ROOT=quote(os.getcwd()),
                    BOB=quote(self.__bobRoot),
                    SPEC=quote(specFile),
                    CLEAN="-c" if cleanWorkspace else "",
                )
        absRunFile = os.path.normpath(os.path.join(workspacePath, runFile))
        absRunFile = os.path.join(".", absRunFile)
        with open(absRunFile, "w") as f:
            print(runFileContent, file=f)
        os.chmod(absRunFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IWGRP |
            stat.S_IROTH | stat.S_IWOTH)

        invoker = Invoker(spec, self.__preserveEnv, self.__noLogFile,
            self.__verbose >= INFO, self.__verbose >= NORMAL,
            self.__verbose >= DEBUG, self.__bufferedStdIO)
        ret = await invoker.executeStep(InvocationMode.CALL, cleanWorkspace)
        if not self.__bufferedStdIO: ttyReinit() # work around MSYS2 messing up the console
        if ret == -int(signal.SIGINT):
            raise BuildError("User aborted while running {}".format(absRunFile),
                             help = "Run again with '--resume' to skip already built packages.")
        elif ret != 0:
            if self.__bufferedStdIO:
                logger.setError(invoker.getStdio().strip())
            raise BuildError("Build script {} returned with {}"
                                .format(absRunFile, ret),
                             help="You may resume at this point with '--resume' after fixing the error.")

    def getStatistic(self):
        return self.__statistic

    def __createGenericTask(self, coro):
        """Create and return task for coroutine."""
        return asyncio.get_event_loop().create_task(self.__taskWrapper(coro))

    def __createCookTask(self, coro, step, checkoutOnly, tracker, count):
        """Create and return task for a cook()-like coroutine.

        The ``step``, ``checkoutOnly`` and the ``tracker`` arguments are used
        to prevent duplicate runs of the same coroutine. Task that only differ
        in the ``checkoutOnly`` parameter wait for each other.

        If ``count`` is True then the task will be counted in the global
        progress.
        """
        sandbox = step.getSandbox() and step.getSandbox().getStep().getVariantId()
        path = (step.getWorkspacePath(), sandbox, checkoutOnly)
        task = tracker.get(path)
        if task is not None: return task

        # Is there a concurrent task running for *not* checkoutOnly? If yes
        # then we have to wait for it to not call the same coroutine
        # concurrently.
        alternatePath = (step.getWorkspacePath(), sandbox, not checkoutOnly)
        alternateTask = tracker.get(alternatePath)

        if count:
            self.__tasksNum += 1
            setProgress(self.__tasksDone, self.__tasksNum)

        task = asyncio.get_event_loop().create_task(self.__taskWrapper(coro,
            path, tracker, alternateTask, step, count))
        tracker[path] = task

        return task

    def __createFingerprintTask(self, coro, step, trackingKey):
        """Create a fingerprinting task identified uniquely by ``trackingKey``.

        The task will be counted in the global progress.
        """
        task = self.__fingerprintTasks.get(trackingKey)
        if task is not None: return task

        self.__tasksNum += 1
        setProgress(self.__tasksDone, self.__tasksNum)

        task = asyncio.get_event_loop().create_task(self.__taskWrapper(coro,
            trackingKey, self.__fingerprintTasks, step=step, count=True))
        self.__fingerprintTasks[trackingKey] = task

        return task

    async def __taskWrapper(self, coro, trackingKey=None, tracker=None,
                            fence=None, step=None, count=False):
        try:
            task = asyncio.Task.current_task()
            self.__allTasks.add(task)
            if fence is not None:
                await fence
            ret = await coro()
            self.__allTasks.remove(task)
            if trackingKey is not None:
                # Only remove us from the task list if we finished successfully.
                # Other concurrent tasks might want to cook the same step again.
                # They have to get the same exception again.
                del tracker[trackingKey]
            if count:
                self.__tasksDone += 1
                setProgress(self.__tasksDone, self.__tasksNum)
            return ret
        except BuildError as e:
            if not self.__keepGoing:
                self.__running = False
            if step:
                e.setStack(step.getPackage().getStack())
            self.__buildErrors.append(e)
            raise CancelBuildException
        except RestartBuildException:
            if self.__running:
                log("Restart build due to wrongly predicted sources.", WARNING)
                self.__restart = True
                self.__running = False
            raise CancelBuildException
        except CancelBuildException:
            raise
        except concurrent.futures.CancelledError:
            raise CancelBuildException
        except Exception as e:
            self.__buildErrors.append(e)
            raise CancelBuildException

    def cook(self, steps, checkoutOnly, depth=0):
        def cancelJobs():
            if self.__jobs > 1:
                log("Cancel all running jobs...", WARNING)
            self.__running = False
            self.__restart = False
            for i in asyncio.Task.all_tasks(): i.cancel()

        async def dispatcher():
            if self.__jobs > 1:
                packageJobs = [
                    self.__createGenericTask(lambda s=step: self._cookTask(s, checkoutOnly, depth))
                    for step in steps ]
                await gatherTasks(packageJobs)
            else:
                packageJobs = []
                for step in steps:
                    job = self.__createGenericTask(lambda s=step: self._cookTask(s, checkoutOnly, depth))
                    packageJobs.append(job)
                    await asyncio.wait({job})
                # retrieve results as last step to --keep-going
                for job in packageJobs: job.result()

        loop = asyncio.get_event_loop()
        self.__restart = True
        while self.__restart:
            self.__running = True
            self.__restart = False
            self.__cookTasks = {}
            self.__buildIdTasks = {}
            self.__fingerprintTasks = {}
            self.__allTasks = set()
            self.__buildErrors = []
            self.__runners = asyncio.BoundedSemaphore(self.__jobs)
            self.__tasksDone = 0
            self.__tasksNum = 0

            j = self.__createGenericTask(dispatcher)
            try:
                loop.add_signal_handler(signal.SIGINT, cancelJobs)
            except NotImplementedError:
                pass # not implemented on windows
            try:
                loop.run_until_complete(j)
            except CancelBuildException:
                pass
            except concurrent.futures.CancelledError:
                pass
            finally:
                try:
                    loop.remove_signal_handler(signal.SIGINT)
                except NotImplementedError:
                    pass # not implemented on windows

            # Reap all remaining tasks to prevent Python warnings about ignored
            # exceptions or that tasks are still pending. We don't care about
            # their result. This is already handled via __buildErrors.
            for i in self.__allTasks: i.cancel()
            if self.__allTasks:
                loop.run_until_complete(asyncio.gather(*self.__allTasks,
                                                       return_exceptions=True))
            self.__allTasks.clear()

            if len(self.__buildErrors) > 1:
                raise MultiBobError(self.__buildErrors)
            elif self.__buildErrors:
                raise self.__buildErrors[0]

        if not self.__running:
            raise BuildError("Canceled by user!",
                             help = "Run again with '--resume' to skip already built packages.")

    async def _cookTask(self, step, checkoutOnly, depth):
        async with self.__runners:
            if not self.__running: raise CancelBuildException
            await self._cook([step], step.getPackage(), checkoutOnly, depth)

    async def _cook(self, steps, parentPackage, checkoutOnly, depth=0):
        # skip everything except the current package
        if self.__skipDeps:
            steps = [ s for s in steps if s.getPackage() == parentPackage ]

        # bail out if nothing has to be done
        steps = [ s for s in steps
                  if s.isValid() and not self._wasAlreadyRun(s, checkoutOnly) ]
        if not steps: return

        if self.__jobs > 1:
            # spawn the child tasks
            tasks = [
                self.__createCookTask(lambda s=step: self._cookStep(s, checkoutOnly, depth),
                                      step, checkoutOnly, self.__cookTasks, True)
                for step in steps
            ]

            # wait for all tasks to finish
            await self.__yieldJobWhile(gatherTasks(tasks))
        else:
            tasks = []
            for step in steps:
                task = self.__createCookTask(lambda s=step: self._cookStep(s, checkoutOnly, depth),
                                             step, checkoutOnly, self.__cookTasks, True)
                tasks.append(task)
                await self.__yieldJobWhile(asyncio.wait({task}))
            # retrieve results as last step to --keep-going
            for t in tasks: t.result()

    async def _cookStep(self, step, checkoutOnly, depth):
        await self.__runners.acquire()
        try:
            if not self.__running:
                raise CancelBuildException
            elif not step.isValid():
                pass
            elif self._wasAlreadyRun(step, checkoutOnly):
                pass
            elif step.isCheckoutStep():
                await self._cook(step.getAllDepSteps(), step.getPackage(), False, depth+1)
                async with self.__workspaceLock(step):
                    if not self._wasAlreadyRun(step, checkoutOnly):
                        await self._cookCheckoutStep(step, depth)
            elif step.isBuildStep():
                await self._cook(step.getAllDepSteps(), step.getPackage(), checkoutOnly, depth+1)
                async with self.__workspaceLock(step):
                    if not self._wasAlreadyRun(step, checkoutOnly):
                        await self._cookBuildStep(step, checkoutOnly, depth)
                        self._setAlreadyRun(step, False, checkoutOnly)
            else:
                assert step.isPackageStep()
                self._preparePackageStep(step)

                # Prohibit up-/download if we are on the old allRelocatable
                # policy and the package is not explicitly relocatable and
                # built outside the sandbox.
                mayUpOrDownload = self.__recipes.getPolicy('allRelocatable') or \
                    step.isRelocatable() or (step.getSandbox() is not None)

                # Calculate build-id and fingerprint of expected artifact if
                # needed. Must be done without the workspace lock because it
                # recurses.
                if checkoutOnly:
                    buildId = None
                else:
                    buildId = await self._getBuildId(step, depth)

                # Try to download if possible. Will only be tried once per
                # invocation!
                downloaded = False
                if mayUpOrDownload and not checkoutOnly:
                    async with self.__workspaceLock(step):
                        if not self._wasDownloadTried(step):
                            downloaded = await self._downloadPackage(step, depth, buildId)
                            self._setDownloadTried(step)
                            if downloaded:
                                self._setAlreadyRun(step, False, checkoutOnly)

                # Recurse and build if not downloaded
                if not downloaded:
                    await self._cook(step.getAllDepSteps(), step.getPackage(), checkoutOnly, depth+1)
                    async with self.__workspaceLock(step):
                        if not self._wasAlreadyRun(step, checkoutOnly):
                            await self._cookPackageStep(step, checkoutOnly, depth, mayUpOrDownload, buildId)
                            self._setAlreadyRun(step, False, checkoutOnly)
        except BuildError as e:
            e.setStack(step.getPackage().getStack())
            raise
        finally:
            # we're done, let the others do their work
            self.__runners.release()

    async def _cookCheckoutStep(self, checkoutStep, depth):
        overrides = set()
        scmList = checkoutStep.getScmList()
        for scm in scmList:
            overrides.update(scm.getActiveOverrides())
        self.__statistic.addOverrides(overrides)
        overrides = len(overrides)
        overridesString = ("(" + str(overrides) + " scm " + ("overrides" if overrides > 1 else "override") +")") if overrides else ""

        # get directory into shape
        (prettySrcPath, created) = self._constructDir(checkoutStep, "src")
        oldCheckoutState = BobState().getDirectoryState(prettySrcPath, True)
        if created:
            # invalidate result if folder was created
            oldCheckoutState = {}
            BobState().resetWorkspaceState(prettySrcPath, oldCheckoutState)
        oldCheckoutHash = BobState().getResultHash(prettySrcPath)

        checkoutExecuted = False
        checkoutDigest = checkoutStep.getVariantId()
        checkoutState = checkoutStep.getScmDirectories().copy()
        checkoutState[None] = (checkoutDigest, None)
        if self.__buildOnly and (BobState().getResultHash(prettySrcPath) is not None):
            if not compareDirectoryState(checkoutState, oldCheckoutState):
                stepMessage(checkoutStep, "CHECKOUT", "WARNING: recipe changed but skipped due to --build-only ({})"
                    .format(prettySrcPath), WARNING)
            else:
                stepMessage(checkoutStep, "CHECKOUT", "skipped due to --build-only ({}) {}".format(prettySrcPath, overridesString),
                    SKIPPED, IMPORTANT)
        else:
            if self.__cleanCheckout:
                # check state of SCMs and invalidate if the directory is dirty
                stats = {}
                for scm in checkoutStep.getScmList():
                    stats[scm.getDirectory()] = scm
                for (scmDir, (scmDigest, scmSpec)) in oldCheckoutState.copy().items():
                    if scmDir is None: continue
                    if scmDigest != checkoutState.get(scmDir, (None, None))[0]: continue
                    if not os.path.exists(os.path.join(prettySrcPath, scmDir)): continue
                    if stats[scmDir].status(checkoutStep.getWorkspacePath()).dirty:
                        # Invalidate scmDigest to forcibly move it away in the loop below.
                        # Do not use None here to distinguish it from a non-existent directory.
                        oldCheckoutState[scmDir] = (False, scmSpec)

            checkoutInputHashes = [ BobState().getResultHash(i.getWorkspacePath())
                for i in checkoutStep.getAllDepSteps() if i.isValid() ]
            if (self.__force or (not checkoutStep.isDeterministic()) or
                (BobState().getResultHash(prettySrcPath) is None) or
                not compareDirectoryState(checkoutState, oldCheckoutState) or
                (checkoutInputHashes != BobState().getInputHashes(prettySrcPath))):
                # move away old or changed source directories
                for (scmDir, (scmDigest, scmSpec)) in oldCheckoutState.copy().items():
                    if (scmDir is not None) and (scmDigest != checkoutState.get(scmDir, (None, None))[0]):
                        scmPath = os.path.normpath(os.path.join(prettySrcPath, scmDir))
                        if os.path.exists(scmPath):
                            atticName = datetime.datetime.now().isoformat().translate(INVALID_CHAR_TRANS)+"_"+os.path.basename(scmPath)
                            stepMessage(checkoutStep, "ATTIC",
                                "{} (move to ../attic/{})".format(scmPath, atticName), WARNING)
                            atticPath = os.path.join(prettySrcPath, "..", "attic")
                            if not os.path.isdir(atticPath):
                                os.makedirs(atticPath)
                            atticPath = os.path.join(atticPath, atticName)
                            os.rename(scmPath, atticPath)
                            BobState().setAtticDirectoryState(atticPath, scmSpec)
                        del oldCheckoutState[scmDir]
                        BobState().setDirectoryState(prettySrcPath, oldCheckoutState)

                # Check that new checkouts do not collide with old stuff in
                # workspace. Do it before we store the new SCM state to
                # check again if the step is rerun.
                for scmDir in checkoutState.keys():
                    if scmDir is None or scmDir == ".": continue
                    if scmDir in oldCheckoutState: continue
                    scmPath = os.path.normpath(os.path.join(prettySrcPath, scmDir))
                    if os.path.exists(scmPath):
                        raise BuildError("New SCM checkout '{}' collides with existing file in workspace '{}'!"
                                            .format(scmDir, prettySrcPath))

                # Store new SCM checkout state. The script state is not stored
                # so that this step will run again if it fails. OTOH we must
                # record the SCM directories as some checkouts might already
                # succeeded before the step ultimately fails.
                BobState().setDirectoryState(prettySrcPath,
                    { d:s for (d,s) in checkoutState.items() if d is not None })

                # Forge checkout result before we run the step again.
                # Normally the correct result is set directly after the
                # checkout finished. But if the step fails and the user
                # re-runs with "build-only" the dependent steps should
                # trigger.
                if BobState().getResultHash(prettySrcPath) is not None:
                    BobState().setResultHash(prettySrcPath, datetime.datetime.utcnow())

                with stepExec(checkoutStep, "CHECKOUT",
                              "{} {}".format(prettySrcPath, overridesString)) as a:
                    await self._runShell(checkoutStep, "checkout", a)
                self.__statistic.checkouts += 1
                checkoutExecuted = True
                # reflect new checkout state
                BobState().setDirectoryState(prettySrcPath, checkoutState)
                BobState().setInputHashes(prettySrcPath, checkoutInputHashes)
                BobState().setVariantId(prettySrcPath, self.__getIncrementalVariantId(checkoutStep))
            else:
                stepMessage(checkoutStep, "CHECKOUT", "skipped (fixed package {})".format(prettySrcPath),
                    SKIPPED, IMPORTANT)

        # We always have to rehash the directory as the user might have
        # changed the source code manually.
        checkoutHash = hashWorkspace(checkoutStep)
        BobState().setResultHash(prettySrcPath, checkoutHash)

        # Generate audit trail. Has to be done _after_ setResultHash()
        # because the result is needed to calculate the buildId.
        if checkoutHash != oldCheckoutHash or self.__force:
            await self._generateAudit(checkoutStep, depth, checkoutHash, checkoutExecuted)

        # upload live build-id cache in case of fresh checkout
        if created and self.__archive.canUploadLocal() and checkoutStep.hasLiveBuildId():
            liveBId = checkoutStep.calcLiveBuildId()
            if liveBId is not None:
                await self.__archive.uploadLocalLiveBuildId(checkoutStep, liveBId, checkoutHash)

        # We're done. The sanity check below won't change the result but would
        # trigger this step again.
        self._setAlreadyRun(checkoutStep, True)

        # Predicted build-id and real one after checkout do not need to
        # match necessarily. Handle it as some build results might be
        # inconsistent to the sources now.
        buildId, predicted = self.__srcBuildIds.get((prettySrcPath, checkoutDigest),
            (checkoutHash, False))
        if buildId != checkoutHash:
            assert predicted, "Non-predicted incorrect Build-Id found!"
            self.__handleChangedBuildId(checkoutStep, checkoutHash)

    async def _cookBuildStep(self, buildStep, checkoutOnly, depth):
        # Add the execution path of the build step to the buildDigest to
        # detect changes between sandbox and non-sandbox builds. This is
        # necessary in any build mode. Include the actual directories of
        # dependencies in buildDigest too. Directories are reused in
        # develop build mode and thus might change even though the variant
        # id of this step is stable. As most tools rely on stable input
        # directories we have to make a clean build if any of the
        # dependency directories change.
        buildDigest = [self.__getIncrementalVariantId(buildStep), buildStep.getExecPath()] + \
            [ i.getExecPath(buildStep) for i in buildStep.getArguments() if i.isValid() ]

        # get directory into shape
        (prettyBuildPath, created) = self._constructDir(buildStep, "build")
        oldBuildDigest = BobState().getDirectoryState(prettyBuildPath, False)
        if created or (buildDigest != oldBuildDigest):
            # not created but exists -> something different -> prune workspace
            if not created and os.path.exists(prettyBuildPath):
                stepMessage(buildStep, "PRUNE", "{} (recipe changed)".format(prettyBuildPath),
                    WARNING)
                emptyDirectory(prettyBuildPath)
            # invalidate build step
            BobState().resetWorkspaceState(prettyBuildPath, buildDigest)

        # run build if input has changed
        buildInputHashes = [ BobState().getResultHash(i.getWorkspacePath())
            for i in buildStep.getAllDepSteps() if i.isValid() ]
        buildFingerprint = await self._getFingerprint(buildStep, depth)
        if buildFingerprint: buildInputHashes.append(buildFingerprint)
        if checkoutOnly:
            stepMessage(buildStep, "BUILD", "skipped due to --checkout-only ({})".format(prettyBuildPath),
                    SKIPPED, IMPORTANT)
        elif (not self.__force) and (BobState().getInputHashes(prettyBuildPath) == buildInputHashes):
            stepMessage(buildStep, "BUILD", "skipped (unchanged input for {})".format(prettyBuildPath),
                    SKIPPED, IMPORTANT)
            # We always rehash the directory in development mode as the
            # user might have compiled the package manually.
            if not self.__cleanBuild:
                BobState().setResultHash(prettyBuildPath, hashWorkspace(buildStep))
        else:
            with stepExec(buildStep, "BUILD", prettyBuildPath) as a:
                # Squash state because running the step will change the
                # content. If the execution fails we have nothing reliable
                # left and we _must_ run it again.
                BobState().delInputHashes(prettyBuildPath)
                BobState().setResultHash(prettyBuildPath, datetime.datetime.utcnow())
                # build it
                await self._runShell(buildStep, "build", a, self.__cleanBuild)
                buildHash = hashWorkspace(buildStep)
            await self._generateAudit(buildStep, depth, buildHash)
            BobState().setResultHash(prettyBuildPath, buildHash)
            BobState().setVariantId(prettyBuildPath, buildDigest[0])
            BobState().setInputHashes(prettyBuildPath, buildInputHashes)

    def _preparePackageStep(self, packageStep):
        # get directory into shape
        (prettyPackagePath, created) = self._constructDir(packageStep, "dist")
        packageDigest = packageStep.getVariantId()
        oldPackageDigest = BobState().getDirectoryState(prettyPackagePath, False)
        if created or (packageDigest != oldPackageDigest):
            # not created but exists -> something different -> prune workspace
            if not created and os.path.exists(prettyPackagePath):
                stepMessage(packageStep, "PRUNE", "{} (recipe changed)".format(prettyPackagePath),
                    WARNING)
                emptyDirectory(prettyPackagePath)
            # invalidate result if folder was created
            BobState().resetWorkspaceState(prettyPackagePath, packageDigest)

    async def _downloadPackage(self, packageStep, depth, packageBuildId):
        # Dissect input parameters that lead to current workspace the last time
        prettyPackagePath = packageStep.getWorkspacePath()
        oldWasDownloaded, oldInputHashes, oldInputBuildId = \
            dissectPackageInputState(BobState().getInputHashes(prettyPackagePath))

        # If possible try to download the package. If we downloaded the
        # package in the last run we have to make sure that the Build-Id is
        # still the same. The overall behaviour should look like this:
        #
        # new workspace -> try download
        # previously built
        #   still same build-id -> normal build
        #   build-id changed -> prune and try download, fall back to build
        # previously downloaded
        #   still same build-id -> done
        #   build-id changed -> prune and try download, fall back to build
        workspaceChanged = False
        wasDownloaded = False
        packageDigest = packageStep.getVariantId()
        if depth >= self.__downloadDepth or (self.__downloadPackages and
                self.__downloadPackages.search(packageStep.getPackage().getName())):
            # prune directory if we previously downloaded/built something different
            if (oldInputBuildId is not None) and (oldInputBuildId != packageBuildId):
                prune = True
                reason = "build-id changed"
            elif self.__force:
                prune = True
                reason = "build forced"
            else:
                prune = False

            if prune:
                stepMessage(packageStep, "PRUNE", "{} ({})".format(prettyPackagePath,
                    reason), WARNING)
                emptyDirectory(prettyPackagePath)
                BobState().resetWorkspaceState(prettyPackagePath, packageDigest)
                oldInputBuildId = None
                oldInputFingerprint = None
                oldInputHashes = None

            # Try to download the package if the directory is currently
            # empty. If the directory holds a result and was downloaded it
            # we're done.
            if BobState().getResultHash(prettyPackagePath) is None:
                audit = os.path.join(prettyPackagePath, "..", "audit.json.gz")
                wasDownloaded = await self.__archive.downloadPackage(packageStep,
                    packageBuildId, audit, prettyPackagePath)
                if wasDownloaded:
                    self.__statistic.packagesDownloaded += 1
                    BobState().setInputHashes(prettyPackagePath,
                        packageInputDownloaded(packageBuildId))
                    packageHash = hashWorkspace(packageStep)
                    workspaceChanged = True
                    wasDownloaded = True
                elif depth >= self.__downloadDepthForce:
                    raise BuildError("Downloading artifact failed")
            elif oldWasDownloaded:
                stepMessage(packageStep, "PACKAGE", "skipped (already downloaded in {})".format(prettyPackagePath),
                    SKIPPED, IMPORTANT)
                wasDownloaded = True

        # Rehash directory if content was changed
        if workspaceChanged:
            BobState().setResultHash(prettyPackagePath, packageHash)
            BobState().setVariantId(prettyPackagePath, packageDigest)
            if wasDownloaded:
                BobState().setInputHashes(prettyPackagePath,
                    packageInputDownloaded(packageBuildId))

        return wasDownloaded

    async def _cookPackageStep(self, packageStep, checkoutOnly, depth, mayUpOrDownload, packageBuildId):
        # Dissect input parameters that lead to current workspace the last time
        prettyPackagePath = packageStep.getWorkspacePath()
        oldWasDownloaded, oldInputHashes, oldInputBuildId = \
            dissectPackageInputState(BobState().getInputHashes(prettyPackagePath))

        # Take checkout step into account because it is guaranteed to
        # be available and the build step might reference it (think of
        # "make -C" or cross-workspace symlinks.
        workspaceChanged = False
        packageInputs = [ packageStep.getPackage().getCheckoutStep() ]
        packageInputs.extend(packageStep.getAllDepSteps())
        packageInputHashes = [ BobState().getResultHash(i.getWorkspacePath())
            for i in packageInputs if i.isValid() ]
        packageFingerprint = await self._getFingerprint(packageStep, depth)
        if packageFingerprint: packageInputHashes.append(packageFingerprint)

        # Run package step if we have not yet downloaded the package or if
        # downloads are not possible anymore. Even if the package was
        # previously downloaded the oldInputHashes will be None to trigger
        # an actual build.
        if checkoutOnly:
            stepMessage(packageStep, "PACKAGE", "skipped due to --checkout-only ({})".format(prettyPackagePath),
                SKIPPED, IMPORTANT)
        elif (not self.__force) and (oldInputHashes == packageInputHashes):
            stepMessage(packageStep, "PACKAGE", "skipped (unchanged input for {})".format(prettyPackagePath),
                SKIPPED, IMPORTANT)
        else:
            with stepExec(packageStep, "PACKAGE", prettyPackagePath) as a:
                # invalidate result because folder will be cleared
                BobState().delInputHashes(prettyPackagePath)
                BobState().setResultHash(prettyPackagePath, datetime.datetime.utcnow())
                await self._runShell(packageStep, "package", a)
                packageHash = hashWorkspace(packageStep)
                packageDigest = self.__getIncrementalVariantId(packageStep)
                workspaceChanged = True
                self.__statistic.packagesBuilt += 1
            audit = await self._generateAudit(packageStep, depth, packageHash)
            if mayUpOrDownload and self.__archive.canUploadLocal():
                await self.__archive.uploadPackage(packageStep, packageBuildId,
                    audit, prettyPackagePath)

        # Rehash directory if content was changed
        if workspaceChanged:
            BobState().setResultHash(prettyPackagePath, packageHash)
            BobState().setVariantId(prettyPackagePath, packageDigest)
            BobState().setInputHashes(prettyPackagePath,
                packageInputBuilt(packageBuildId, packageInputHashes))

    async def __queryLiveBuildId(self, step):
        """Predict live build-id of checkout step.

        Query the SCMs for their live-buildid and cache the result. Normally
        the current result is retuned unless we're in build-only mode. Then the
        cached result is used. Only if there is no cached entry the query is
        performed.
        """

        key = b'\x00' + step._getSandboxVariantId()
        if self.__buildOnly:
            liveBId = BobState().getBuildId(key)
            if liveBId is not None: return liveBId

        liveBId = await step.predictLiveBuildId()
        if liveBId is not None:
            BobState().setBuildId(key, liveBId)
        return liveBId

    def __invalidateLiveBuildId(self, step):
        """Invalidate last live build-id of a step."""

        key = b'\x00' + step._getSandboxVariantId()
        liveBId = BobState().getBuildId(key)
        if liveBId is not None:
            BobState().delBuildId(key)

    async def __translateLiveBuildId(self, step, liveBId):
        """Translate live build-id into real build-id.

        We maintain a local cache of previous translations. In case of a cache
        miss the archive is interrogated. A valid result is cached.
        """
        key = b'\x01' + liveBId
        bid = BobState().getBuildId(key)
        if bid is not None:
            return bid

        bid = await self.__archive.downloadLocalLiveBuildId(step, liveBId)
        if bid is not None:
            BobState().setBuildId(key, bid)

        return bid

    async def __getCheckoutStepBuildId(self, step, depth):
        ret = None

        # Try to use live build-ids for checkout steps. Do not use them if
        # there is already a workspace or if the package matches one of the
        # 'always-checkout' patterns. Fall back to a regular checkout if any
        # condition is not met.
        name = step.getPackage().getName()
        path = step.getWorkspacePath()
        if not os.path.exists(step.getWorkspacePath()) and \
           not any(pat.search(name) for pat in self.__alwaysCheckout) and \
           step.hasLiveBuildId() and self.__archive.canDownloadLocal():
            with stepAction(step, "QUERY", step.getPackage().getName(), (IMPORTANT, NORMAL)) as a:
                liveBId = await self.__queryLiveBuildId(step)
                if liveBId:
                    ret = await self.__translateLiveBuildId(step, liveBId)
                if ret is None:
                    a.fail("unknown", WARNING)

        # do the checkout if we still don't have a build-id
        if ret is None:
            await self._cook([step], step.getPackage(), False, depth)
            # return directory hash
            ret = BobState().getResultHash(step.getWorkspacePath())
            predicted = False
        else:
            predicted = True

        return ret, predicted

    async def _getBuildId(self, step, depth):
        """Calculate build-id and cache result.

        The cache uses the workspace path as index because there might be
        multiple directories with the same variant-id. As the src build-ids can
        be cached for a long time the variant-id is used as index too to
        prevent possible false hits if the recipes change between runs.

        Checkout steps are cached separately from build and package steps.
        Build-ids of checkout steps may be predicted through live-build-ids. If
        we the prediction was wrong the build and package step build-ids are
        invalidated because they could be derived from the wrong checkout
        build-id.
        """

        # Pass over to __getBuildIdList(). It will try to create a task and (if
        # the calculation already failed) will make sure that we get the same
        # exception again. This prevents recalculation of already failed build
        # ids.
        [ret] = await self.__getBuildIdList([step], depth)
        return ret

    async def __getBuildIdList(self, steps, depth):
        if self.__jobs > 1:
            tasks = [
                self.__createCookTask(lambda s=step: self.__getBuildIdTask(s, depth),
                                      step, False, self.__buildIdTasks, False)
                for step in steps
            ]
            ret = await self.__yieldJobWhile(gatherTasks(tasks), True)
        else:
            tasks = []
            for step in steps:
                task = self.__createCookTask(lambda s=step: self.__getBuildIdTask(s, depth),
                                             step, False, self.__buildIdTasks, False)
                tasks.append(task)
                await self.__yieldJobWhile(asyncio.wait({task}), True)
            # retrieve results as last step to --keep-going
            ret = [ t.result() for t in tasks ]
        return ret

    async def __getBuildIdTask(self, step, depth):
        async with self.__runners:
            ret = await self.__getBuildIdSingle(step, depth)
        return ret

    async def __getBuildIdSingle(self, step, depth):
        path = step.getWorkspacePath()
        if step.isCheckoutStep():
            key = (path, step.getVariantId())
            ret, predicted = self.__srcBuildIds.get(key, (None, False))
            if ret is None:
                tmp = await self.__getCheckoutStepBuildId(step, depth)
                self.__srcBuildIds[key] = tmp
                ret = tmp[0]
        else:
            ret = self.__buildDistBuildIds.get(path)
            if ret is None:
                fingerprint = await self._getFingerprint(step, depth)
                ret = await step.getDigestCoro(lambda x: self.__getBuildIdList(x, depth+1),
                    True, fingerprint=fingerprint, platform=getPlatformTag())
                self.__buildDistBuildIds[path] = ret

        return ret

    def __handleChangedBuildId(self, step, checkoutHash):
        """Handle different build-id of src step after checkout.

        Through live-build-ids it is possible that an initially queried
        build-id does not match the real build-id after the sources have been
        checked out. As we might have already downloaded artifacts based on
        the now invalidated build-id we have to restart the build and check all
        build-ids, build- and package-steps again.
        """
        key = (step.getWorkspacePath(), step.getVariantId())

        # Invalidate wrong live-build-id
        self.__invalidateLiveBuildId(step)

        # Invalidate (possibly) derived build-ids
        self.__srcBuildIds[key] = (checkoutHash, False)
        self.__buildDistBuildIds = {}

        # Forget all executed build- and package-steps
        self._clearWasRun()
        self._clearDownloadTried()

        # start from scratch
        raise RestartBuildException()

    def __getIncrementalVariantId(self, step):
        """Calculate the variant-id with respect to workspace state.

        The real variant-id can be calculated solely by looking at the recipes.
        But as we allow the user to build single packages, skip dependencies
        and support checkout-/build-only builds the actual variant-id is
        dependent on the current project state.

        For every workspace we store the variant-id of the last build. When
        calculating the incremental variant-id of a step we take these stored
        variant-ids for the dependencies. If no variant-id was stored we take
        the real one because this is typically an old workspace where we want
        to prevent useless rebuilds. It could also be that the workspace was
        deleted.

        Important: this method can only work reliably if the dependent steps
        have been cooked. Otherwise the state may have stale data.
        """

        def getStoredVId(dep):
            ret = BobState().getVariantId(dep.getWorkspacePath())
            if ret is None:
                ret = dep.getVariantId()
            return ret

        return step.getDigest(getStoredVId)

    async def __yieldJobWhile(self, coro, ignoreExecutionStop = False):
        """Yield the job slot while waiting for a coroutine.

        Handles the dirty details of cancellation. Might throw CancelledError
        if overall execution was stopped.
        """
        self.__runners.release()
        try:
            ret = await coro
        finally:
            acquired = False
            while not acquired:
                try:
                    await self.__runners.acquire()
                    acquired = True
                except concurrent.futures.CancelledError:
                    pass
        if not self.__running and not ignoreExecutionStop: raise CancelBuildException
        return ret

    async def _getFingerprint(self, step, depth):
        # Use a shortcut when the sandboxFingerprints policy is not set and the
        # step is built inside a sandbox. In this case the variant-id and
        # build-id are already tied directly to the sandbox variant-id by
        # getDigest(). This is pessimistic but easier than calculating the
        # fingerprint inside the sandbox and has been the default before Bob
        # 0.16.
        sandbox = (step.getSandbox() is not None) and step.getSandbox().getStep()
        if sandbox and not self.__recipes.getPolicy('sandboxFingerprints'):
            return b''

        # A relocatable step with no fingerprinting is easy
        isFingerprinted = step._isFingerprinted()
        trackRelocation = step.isPackageStep() and not step.isRelocatable() and \
            self.__recipes.getPolicy('allRelocatable')
        if not isFingerprinted and not trackRelocation:
            return b''

        # Execute the fingerprint script (or use cached result)
        if isFingerprinted:
            # Cache based on the script digest.
            key = hashlib.sha1(step._getFingerprintScript().encode('utf8')).digest()
            if sandbox:
                # Add the sandbox build-id to the cache key to distinguish
                # between different sandboxes.
                sandboxBuildId = await self._getBuildId(sandbox, depth+1)
                key = hashlib.sha1(key + sandboxBuildId).digest()

            # Run fingerprint calculation in another task. The task will only
            # be created once for each ``key``.
            fingerprint = self.__fingerprints.get(key)
            if fingerprint is None:
                fingerprintTask = self.__createFingerprintTask(
                    lambda: self.__calcFingerprintTask(step, sandbox, key, depth),
                    step, key)
                await self.__yieldJobWhile(asyncio.wait({fingerprintTask}), True)
                fingerprint = fingerprintTask.result()
        else:
            fingerprint = b''

        # If the package is not relocatable the exec path is mixed into the
        # fingerprint to tag the relocation information at the artifact.
        if trackRelocation:
            fingerprint += step.getExecPath().encode(
                locale.getpreferredencoding(False), 'replace')

        return hashlib.sha1(fingerprint).digest()

    async def __calcFingerprintTask(self, step, sandbox, key, depth):
        async with self.__runners:
            # If this is built in a sandbox then the artifact cache may help...
            if sandbox:
                fingerprint = await self.__archive.downloadLocalFingerprint(sandbox, key)
            else:
                fingerprint = None

            # When we don't know it yet we really have to execute the script.
            # In case a sandbox is used we have to make sure it's available.
            if fingerprint is None:
                if sandbox:
                    await self._cook([sandbox], sandbox.getPackage(), False, depth+1)
                with stepAction(step, "FNGRPRNT", step.getPackage().getName()) as a:
                    fingerprint = await self.__runFingerprintScript(step, a)

                # Always upload if this was calculated in a sandbox. The task will
                # only be run once so we don't need to worry here about duplicate
                # uploads.
                if sandbox:
                    await self.__archive.uploadLocalFingerprint(sandbox, key, fingerprint)

            # Cache result so that we don't ever need to spawn a task
            self.__fingerprints[key] = fingerprint

        return fingerprint

    async def __runFingerprintScript(self, step, logger):
        spec = StepSpec.fromStep(step, None, self.__envWhiteList)
        invoker = Invoker(spec, self.__preserveEnv, True, True, True, False, True)
        (ret, stdout, stderr) = await invoker.executeFingerprint()

        if ret == -int(signal.SIGINT):
            raise BuildError("Fingerprint script interrupted by user")
        elif ret != 0:
            help = "Script output: " + stderr.decode(
                locale.getpreferredencoding(False), 'replace').strip()
            if self.__verbose >= DEBUG:
                help += "\nScript: " + spec.fingerprintScript
            raise BuildError("Fingerprint script returned with {}".format(ret),
                help=help)

        log = repr(stdout.decode(locale.getpreferredencoding(False), 'replace').strip())
        if len(log) > 43: log = log[:40] + "..."
        logger.setResult(log)
        return stdout



# Bob build tool
# Copyright (C) 2017  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .errors import BobError
from .stringparser import isFalse, isTrue, Env
from .utils import infixBinaryOp
from collections import OrderedDict
from itertools import chain
from fnmatch import fnmatchcase
from functools import lru_cache
import pickle
import pyparsing
import sqlite3

# need to enable this for nested expression parsing performance
pyparsing.ParserElement.enablePackrat()

# See "Efficient algorithms for processing XPath queries" [1] for the core
# algorithms that are applied here.
#
# [1] Gottlob, G., Koch, C., and Pichler, R. 2002. Efficient algorithms
#     for processing XPath queries. In Proceedings of the 28th
#     International Conference on Very Large Data Bases (VLDB'02).
#     HongKong, China, 95-106


class SandboxSentinel:
    def isEnabled(self):
        return False
sandboxSentinel = SandboxSentinel()

def markLocation(line, loc):
    if len(line) <= 60:
        # line is short enough
        pass
    elif loc <= 30:
        # error at beginning
        line = line[:60] + '[...]'
    elif loc >= (len(line) - 30):
        # error at end
        loc = 60 - (len(line) - loc) + 5
        line = '[...]' + line[-60:]
    else:
        # error in middle
        line = '[...]' + line[loc-30:loc+30] + '[...]'
        loc = 35

    return "Offending query: " + line + "\n" + (" " * (loc + 17)) + \
        "^.-- Error location"

class BaseASTNode:
    def __init__(self, s, loc):
        self.__s = s
        self.__loc = loc

    def barf(self, msg):
        raise BobError("Bad syntax: " + msg + " (at char {})".format(self.__loc),
                       help=markLocation(self.__s, self.__loc))

class LocationPath(BaseASTNode):
    """AST class that represents a 'location path'"""

    def __init__(self, s, loc, toks, rootNodeGetter):
        super().__init__(s, loc)
        self.__path = [
            (LocationStep(s, loc, ['descendant-or-self', '@', '*']) if t == "//" else t)
            for t in toks
            if t != "/"
        ]
        self.__absolute = (toks[0] == '/') or (toks[0] == '//')
        self.__rootNodeGetter = rootNodeGetter

        # remove trivial 'self' steps
        self.__path = [ s for s in self.__path
            if (s.axis != 'self' or s.test != '*' or s.pred is not None) ]

        # combine '//foo' to 'descendant@foo'
        # FIXME: this looks ugly as hell. there's certainly a better way
        path = []
        i = 0
        while i < len(self.__path):
            step = self.__path[i]
            if i < len(self.__path)-1:
                nextStep = self.__path[i+1]
                if (step.axis == 'descendant-or-self' and step.test == '*' and
                        step.pred is None and nextStep.axis == 'child'):
                    nextStep.axis = 'descendant'
                    step = nextStep
                    i += 1
            path.append(step)
            i += 1
        self.__path = path

    def __repr__(self):
        return "LocationPath({})".format(self.__path)

    def __findIntermediateNodes(self, old, new, queryIndirect):
        """Find nodes that are on on any path between 'old' and 'new'"""

        visited = set()
        intermediate = set()
        if old.issuperset(new): return intermediate

        def traverse(node, stack):
            if node in visited: return

            if node in new:
                intermediate.update(stack)
            else:
                stack = stack + [node]
                for i in node.values():
                    if queryIndirect or i.direct:
                        traverse(i.node, stack)
                visited.add(node)

        for n in old: traverse(n, [])

        return intermediate

    def __findReachableSubset(self, valid, nodes):
        """Find all nodes and their parents that are in the valid set."""

        ret = set()
        todo = set(nodes)
        while todo:
            node = todo.pop()
            if (node not in valid) or (node in ret): continue
            ret.add(node)
            todo.update(node.parents(True))

        return ret

    def evalForward(self, root):
        """Evaluate path step by step.

        Each step is performed with the set of context nodes that represent the
        (intermediate) result.  While each step is evaluated the trail of
        visited nodes is recorded in 'valid'. After each step these nodes are
        trimmed by the parent-reachable nodes of the current context nodes.

        The 'valid' set is returned to limit the possible paths of the result
        nodes to what was visited during the path traversal. Even though any
        path would technically be correct the user intuitively expects the
        resulting paths to reflect the intermediate steps of the query.
        """
        nodes = set([root])
        valid = set([root])
        for i in self.__path:
            oldNodes = nodes
            nodes, search = i.evalForward(nodes, valid)
            if search is not None:
                valid.update(self.__findIntermediateNodes(oldNodes, nodes, search))
            else:
                valid.update(nodes)
            valid.update(nodes)
            valid.intersection_update(self.__findReachableSubset(valid, nodes))
        return (nodes, valid)

    def evalBackward(self):
        """Evaluate a path backwards.

        Used in predicate expressions to find all nodes that possibly match the
        path. IOW we backwards find all nodes that satisfy the path.
        """
        root = self.__rootNodeGetter()
        allNodes = nodes = set(root.allNodes())
        for i in reversed(self.__path):
            nodes = i.evalBackward(nodes)
        if self.__absolute:
            if root in nodes:
                nodes = allNodes
            else:
                nodes = set()
        return nodes

    def evalString(self, pkg):
        """A path cannot be treated as a string"""
        self.barf("location path in string context")

class LocationStep(BaseASTNode):
    """AST class that represents a single step in a location path."""

    def __init__(self, s, loc, step):
        """Constructor.

        Handles the various abbreviations of steps like the optional axis, the
        optional predicate and the '.' short-cut.
        """
        super().__init__(s, loc)
        self.__pred = None
        if len(step) == 1:
            if step[0] == '.':
                self.__axis = 'self'
                self.__test = '*'
            else:
                self.__axis = 'child'
                self.__test = step[0]
        else:
            if step[1] == '@':
                self.__axis = step[0]
                self.__test = step[2]
                remain = step[3:]
            else:
                self.__axis = 'child'
                self.__test = step[0]
                remain = step[1:]

            if remain:
                assert remain[0] == '['
                assert remain[2] == ']'
                self.__pred = remain[1]

    def __repr__(self):
        return "LocationStep({}@{}[{}])".format(self.__axis, self.__test, self.__pred)

    def __evalAxisChild(self, nodes, queryIndirect):
        """Find nodes in the 'child' axis."""
        ret = set()
        for i in nodes:
            ret.update(c.node for c in i.values() if (queryIndirect or c.direct))
        return ret

    def __evalAxisDescendant(self, nodes, queryIndirect):
        """Find nodes in the 'descendant' axis."""
        ret = set()
        todo = nodes
        while todo:
            childs = set()
            for i in todo:
                childs.update(c.node for c in i.values()
                              if (queryIndirect or c.direct))
            todo = childs - ret
            ret.update(childs)
        return ret

    def __evalAxisParent(self, nodes, queryIndirect):
        """Find nodes in the 'parent' axis."""
        ret = set()
        for i in nodes:
            ret.update(i.parents(queryIndirect))
        return ret

    def __evalAxisAncestor(self, nodes, queryIndirect):
        """Find nodes in the 'ancestor' axis."""
        ret = set()
        todo = nodes
        while todo:
            parents = set()
            for i in todo: parents.update(i.parents(queryIndirect))
            todo = parents - ret
            ret.update(parents)
        return ret

    @property
    def axis(self):
        return self.__axis

    @axis.setter
    def axis(self, value):
        assert value in ['self', 'child', 'descendant', 'descendant-or-self'], value
        self.__axis = value

    @property
    def test(self):
        return self.__test

    @property
    def pred(self):
        return self.__pred

    def evalForward(self, nodes, valid):
        """Evaluate the axis, name test and predicate

        Despite the result set returns whether we possibly made multiple hops
        in the dendency graph, i.e.  evaluated a 'descendant' axis. In this
        caste it is the responsibility of the caller to calculate all possible
        paths that lead to the result set.
        """
        search = None
        if self.__axis == "child":
            nodes = self.__evalAxisChild(nodes, True)
        elif self.__axis == "descendant":
            nodes = self.__evalAxisDescendant(nodes, True)
            search = True
        elif self.__axis == "descendant-or-self":
            nodes = self.__evalAxisDescendant(nodes, True) | nodes
            search = True
        elif self.__axis == "direct-child":
            nodes = self.__evalAxisChild(nodes, False)
        elif self.__axis == "direct-descendant":
            nodes = self.__evalAxisDescendant(nodes, False)
            search = False
        elif self.__axis == "direct-descendant-or-self":
            nodes = self.__evalAxisDescendant(nodes, False) | nodes
            search = False
        elif self.__axis == "self":
            pass
        else:
            assert False, "Invalid axis: " + self.__axis

        if self.__test == "*":
            pass
        elif '*' in self.__test:
            nodes = set(i for i in nodes if fnmatchcase(i.getName(), self.__test))
        else:
            nodes = set(i for i in nodes if i.getName() == self.__test)

        if self.__pred:
            nodes = nodes & self.__pred.evalBackward()

        return (nodes, search)

    def evalBackward(self, nodes):
        """Inverse evaluation of location path step."""
        if self.__test == "*":
            pass
        elif '*' in self.__test:
            nodes = set(i for i in nodes if fnmatchcase(i.getName(), self.__test))
        else:
            nodes = set(i for i in nodes if i.getName() == self.__test)

        if self.__pred:
            nodes = nodes & self.__pred.evalBackward()

        if self.__axis == "child":
            nodes = self.__evalAxisParent(nodes, True)
        elif self.__axis == "descendant":
            nodes = self.__evalAxisAncestor(nodes, True)
        elif self.__axis == "descendant-or-self":
            nodes = self.__evalAxisAncestor(nodes, True) | nodes
        elif self.__axis == "direct-child":
            nodes = self.__evalAxisParent(nodes, False)
        elif self.__axis == "direct-descendant":
            nodes = self.__evalAxisAncestor(nodes, False)
        elif self.__axis == "direct-descendant-or-self":
            nodes = self.__evalAxisAncestor(nodes, False) | nodes
        elif self.__axis == "self":
            pass
        else:
            assert False, "Invalid axis: " + self.__axis

        return nodes

class NotOperator(BaseASTNode):
    def __init__(self, s, loc, toks, rootNodeGetter):
        super().__init__(s, loc)
        self.rootNodeGetter = rootNodeGetter
        assert len(toks) == 1, toks
        toks = toks[0]
        assert len(toks) == 2, toks
        assert toks[0] == '!'
        self.op = toks[1]

    def __repr__(self):
        return "NotOperator({})".format(self.op)

    def evalBackward(self):
        return set(self.rootNodeGetter().allNodes()) - self.op.evalBackward()

    def evalString(self, pkg):
        self.barf("operator in string context")

class BinaryBoolOperator(BaseASTNode):
    def __init__(self, s, loc, toks):
        super().__init__(s, loc)
        self.left = toks[0]
        self.right = toks[2]
        self.opStr = op = toks[1]
        if op == '&&':
            self.op = lambda l, r: l & r
        elif op == '||':
            self.op = lambda l, r: l | r
        else:
            assert False, op

    def __repr__(self):
        return "BinaryBoolOperator({}, {}, {})".format(self.left, self.opStr, self.right)

    def evalBackward(self):
        return self.op(self.left.evalBackward(), self.right.evalBackward())

    def evalString(self, pkg):
        self.barf("operator in string context")

class StringLiteral(BaseASTNode):
    def __init__(self, s, loc, toks, doSubst, stringFunctions, graphIterator):
        super().__init__(s, loc)
        assert len(toks) == 1, toks
        self.literal = toks[0]
        self.subst = doSubst and any((c in self.literal) for c in '\\\"\'$')
        self.stringFunctions = stringFunctions
        self.graphIterator = graphIterator

    def __repr__(self):
        if self.subst:
            return "StringLiteral(\"{}\")".format(self.literal)
        else:
            return "StringLiteral('{}')".format(self.literal)

    def evalBackward(self):
        return set( n for (n, p) in self.graphIterator()
                    if isTrue(self.evalString(p)) )

    def evalString(self, pkg):
        if self.subst:
            pkgStep = pkg.getPackageStep()
            env = Env(pkgStep.getEnv())
            env.update(pkg.getMetaEnv())
            env.setFunArgs({
                "package" : pkg,
                "recipe" : pkg.getRecipe(),
                "sandbox" : (pkg._getSandboxRaw() or sandboxSentinel).isEnabled(),
                "__tools" : pkgStep.getTools()
            })
            env.setFuns(self.stringFunctions)
            return env.substitute(self.literal, self.literal, False)
        else:
            return self.literal

class FunctionCall(BaseASTNode):
    def __init__(self, s, loc, toks, stringFunctions, graphIterator):
        super().__init__(s, loc)
        if toks[0] not in stringFunctions:
            self.barf("Unknown string function: " + toks[0])
        self.fun = stringFunctions[toks[0]]
        self.name = toks[0]
        self.args = toks[1:]
        self.graphIterator = graphIterator

    def __repr__(self):
        return "FunctionCall({}, {})".format(self.name,
            ", ".join(repr(a) for a in self.args))

    def evalBackward(self):
        return set( n for (n, p) in self.graphIterator()
                    if isTrue(self.evalString(p)) )

    def evalString(self, pkg):
        pkgStep = pkg.getPackageStep()
        env = Env(pkgStep.getEnv())
        env.update(pkg.getMetaEnv())
        args = [ a.evalString(pkg) for a in self.args ]
        extra = {
            "package" : pkg,
            "recipe" : pkg.getRecipe(),
            "sandbox" : (pkg._getSandboxRaw() or sandboxSentinel).isEnabled(),
            "__tools" : pkgStep.getTools()
        }
        return self.fun(args, env=env, **extra)

class BinaryStrOperator(BaseASTNode):
    def __init__(self, s, loc, toks, graphIterator):
        super().__init__(s, loc)
        self.left = toks[0]
        self.right = toks[2]
        self.opStr = op = toks[1]
        if op == '<':
            self.op = lambda l, r: l < r
        elif op == '>':
            self.op = lambda l, r: l > r
        elif op == '<=':
            self.op = lambda l, r: l <= r
        elif op == '>=':
            self.op = lambda l, r: l >= r
        elif op == '==':
            self.op = lambda l, r: l == r
        elif op == '!=':
            self.op = lambda l, r: l != r
        else:
            assert False, op
        self.graphIterator = graphIterator

    def __repr__(self):
        return "BinaryStrOperator({}, {}, {})".format(self.left,
            self.opStr, self.right)

    def evalBackward(self):
        return set( n for (n, p) in self.graphIterator()
                    if self.op(self.left.evalString(p), self.right.evalString(p)) )

    def evalString(self, pkg):
        self.barf("operator in string context")


class PkgGraphEdge:
    __slots__ = ['__node', '__isDirect', '__origin']

    def __init__(self, db, data):
        self.__node = PkgGraphNode(db, data[0])
        self.__isDirect = data[1]
        self.__origin = data[2]

    @property
    def node(self):
        return self.__node

    @property
    def direct(self):
        return self.__isDirect

    @property
    def origin(self):
        return self.__origin

class PkgGraphNode:
    __slots__ = ['__db', '__name', '__key', '__parents', '__childs']

    def __init__(self, db, key, node=None):
        self.__db = db
        self.__key = key
        try:
            if node is None:
                db.execute("SELECT node FROM graph WHERE key=?", (key,))
                node = db.fetchone()[0]
            (self.__name, self.__parents, self.__childs) = pickle.loads(node)
        except sqlite3.Error as e:
            raise BobError("Cannot load internal state: " + str(e))

    @classmethod
    def init(cls, cacheName, cacheKey, rootGenerator):
        try:
            db = sqlite3.connect(cacheName, isolation_level=None).cursor()
            db.execute("CREATE TABLE IF NOT EXISTS meta(key PRIMARY KEY, value)")
            db.execute("CREATE TABLE IF NOT EXISTS graph(key PRIMARY KEY, node)")

            # check if Bob was changed
            db.execute("BEGIN")
            db.execute("SELECT value FROM meta WHERE key='vsn'")
            vsn = db.fetchone()
            if (vsn is None) or (vsn[0] != cacheKey):
                # Database was changed or created
                db.execute("DELETE FROM graph")
                root = PkgGraphNode.__convertPackageToGraph(db, rootGenerator())
                db.execute("INSERT OR REPLACE INTO meta VALUES ('vsn', ?), ('root', ?)",
                    (cacheKey, root))
                # Commit and start new read-only transaction
                db.execute("END")
                db.execute("BEGIN")
            else:
                # Database is valid. Use it...
                db.execute("SELECT value FROM meta WHERE key='root'")
                root = db.fetchone()[0]

            return cls(db, root)

        except sqlite3.Error as e:
            raise BobError("Cannot save internal state: " + str(e))

    def close(self):
        self.__db.close()
        self.__db.connection.close()
        del self.__db

    def __repr__(self):
        return "Node({})".format(self.__key)

    def __hash__(self):
        return hash(self.__key)

    def __eq__(self, other):
        return isinstance(other, PkgGraphNode) and (self.__key == other.__key)

    def __len__(self):
        return len(self.__childs)

    def __contains__(self, name):
        return name in self.__childs

    def __getitem__(self, name):
        return PkgGraphEdge(self.__db, self.__childs[name])

    def __iter__(self):
        return iter(self.__childs)

    def key(self):
        return self.__key

    def getByKey(self, key):
        return PkgGraphEdge(self.__db, key)

    def keys(self):
        return self.__childs.keys()

    def values(self):
        return iter( PkgGraphEdge(self.__db, i) for i in self.__childs.values() )

    def items(self):
        return iter( (name, PkgGraphEdge(self.__db, child))
                     for name, child in self.__childs.items() )

    def parents(self, queryIndirect):
        return iter( PkgGraphNode(self.__db, p) for (p, d) in self.__parents.items()
                     if (queryIndirect or d) )

    def allNodes(self):
        self.__db.execute("SELECT key FROM graph")
        keys = self.__db.fetchall()
        return iter( PkgGraphNode(self.__db, i) for (i,) in keys )

    def getName(self):
        return self.__name

    def __addParent(self, parent, direct):
        # Direct dependencies are traversed first. We don't need to worry that
        # a parent is flipping between direct and indirect.
        if parent not in self.__parents:
            self.__parents[parent] = direct
            node = (self.__name, self.__parents, self.__childs)
            self.__db.execute("INSERT OR REPLACE INTO graph VALUES (?, ?)",
                (self.__key, pickle.dumps(node, -1)))

    @staticmethod
    def __convertPackageToGraph(db, pkg, parent=None, directParent=True):
        name = pkg.getName()
        key = pkg._getId()
        db.execute("SELECT node FROM graph WHERE key=?", (key,))
        node = db.fetchone()
        if node is None:
            # recurse
            childs = OrderedDict()
            for d in pkg.getDirectDepSteps():
                subPkg = d.getPackage()
                subPkgId = PkgGraphNode.__convertPackageToGraph(db, subPkg, key, True)
                childs[subPkg.getName()] = (subPkgId, True, "")
            prefixLen = len("/".join(pkg.getStack()))
            for d in pkg.getIndirectDepSteps():
                subPkg = d.getPackage()
                subPkgName = subPkg.getName()
                if subPkgName in childs: continue
                subPkgId = PkgGraphNode.__convertPackageToGraph(db, subPkg, key, False)
                childs[subPkgName] = ( subPkgId, False,
                    ".." + "/".join(subPkg.getStack())[prefixLen:] )
            # create node
            node = ( name, ({parent:directParent} if parent is not None else {}), childs )
            db.execute("INSERT INTO graph VALUES (?, ?)",
                (key, pickle.dumps(node, -1)))
        elif parent is not None:
            # add as parent
            PkgGraphNode(db, key, node[0]).__addParent(parent, directParent)

        return key


class GraphPackageIterator:
    """Special iterator that yields the 'PkgGraphNode' node _and_ the 'Package'
    together.

    The traversal is done in lock step for maximum efficiency. The culprit is
    that the 'bob.input.Package' and 'bob.input.Step' objects are re-created on
    the fly. We must make sure that this is done only once per package.
    """
    def __init__(self, graphRoot, packageRoot):
        self.__graphRoot = graphRoot
        self.__pkgRoot = packageRoot

    def __iter__(self):
        stack = [ (self.__graphRoot, chain(self.__pkgRoot.getDirectDepSteps(),
                                           self.__pkgRoot.getIndirectDepSteps())) ]
        yield (self.__graphRoot, self.__pkgRoot)
        done = set([self.__graphRoot.key()])

        while stack:
            try:
                childPkg = next(stack[-1][1]).getPackage()
                childNode = stack[-1][0][childPkg.getName()].node
                if childNode.key() not in done:
                    done.add(childNode.key())
                    yield (childNode, childPkg)
                    stack.append( (childNode, chain(childPkg.getDirectDepSteps(),
                                                    childPkg.getIndirectDepSteps())) )
            except StopIteration:
                stack.pop()


class PackageSet:
    """Accessor object to the calculated package set.

    Takes care of the transparent caching of package adjacency matrix that is
    used for the path query evaluation.
    """

    def __init__(self, cacheKey, aliases, stringFunctions, packageGenerator):
        self.__cacheKey = cacheKey
        self.__aliases = aliases
        self.__stringFunctions = stringFunctions
        self.__generator = packageGenerator
        self.__root = None
        self.__graph = None

        # create parsing grammer
        locationPath = pyparsing.Forward()
        relativeLocationPath = pyparsing.Forward()

        axisName = \
              pyparsing.Keyword("descendant-or-self") \
            | pyparsing.Keyword("child") \
            | pyparsing.Keyword("descendant") \
            | pyparsing.Keyword("direct-descendant-or-self") \
            | pyparsing.Keyword("direct-child") \
            | pyparsing.Keyword("direct-descendant") \
            | pyparsing.Keyword("self")

        nodeTest = pyparsing.Word(pyparsing.alphanums + "_.:+-*")
        axisSpecifier = axisName + '@'
        abbreviatedStep = pyparsing.Keyword('.')

        sQStringLiteral = pyparsing.QuotedString("'")
        sQStringLiteral.setParseAction(
            lambda s, loc, toks: StringLiteral(s, loc, toks, False,
                                               self.__stringFunctions,
                                               self.__getGraphIter))
        dQStringLiteral = pyparsing.QuotedString('"', '\\')
        dQStringLiteral.setParseAction(
            lambda s, loc, toks: StringLiteral(s, loc, toks, True,
                                               self.__stringFunctions,
                                               self.__getGraphIter))
        stringLiteral = sQStringLiteral | dQStringLiteral

        functionCall = pyparsing.Forward()
        functionArg = stringLiteral | functionCall
        functionCall << pyparsing.Word(pyparsing.alphas, pyparsing.alphanums+'-') + \
            pyparsing.Suppress('(') + \
            pyparsing.Optional(functionArg +
                pyparsing.ZeroOrMore(pyparsing.Suppress(',') + functionArg)) + \
            pyparsing.Suppress(')')
        functionCall.setParseAction(
            lambda s, loc, toks: FunctionCall(s, loc, toks, self.__stringFunctions,
                                              self.__getGraphIter))

        predExpr = pyparsing.infixNotation(
            locationPath ^ stringLiteral ^ functionCall,
            [
                ('!',  1, pyparsing.opAssoc.RIGHT, lambda s, loc, toks: NotOperator(s, loc, toks, self.__getGraphRoot)),
                ('<',  2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryStrOperator, self.__getGraphIter)),
                ('<=', 2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryStrOperator, self.__getGraphIter)),
                ('>',  2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryStrOperator, self.__getGraphIter)),
                ('>=', 2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryStrOperator, self.__getGraphIter)),
                ('==', 2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryStrOperator, self.__getGraphIter)),
                ('!=', 2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryStrOperator, self.__getGraphIter)),
                ('&&', 2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryBoolOperator)),
                ('||', 2, pyparsing.opAssoc.LEFT,  infixBinaryOp(BinaryBoolOperator))
            ])
        predicate = '[' + predExpr + ']'
        step = abbreviatedStep | (pyparsing.Optional(axisSpecifier) +
                                  nodeTest + pyparsing.Optional(predicate))
        step.setParseAction(lambda s, loc, toks: LocationStep(s, loc, toks))
        abbreviatedRelativeLocationPath = step + '//' + relativeLocationPath
        relativeLocationPath << (
            abbreviatedRelativeLocationPath |
            (step + '/' + relativeLocationPath) |
            step)
        abbreviatedAbsoluteLocationPath = '//' + relativeLocationPath
        absoluteLocationPath = abbreviatedAbsoluteLocationPath | \
                               ('/' + relativeLocationPath)
        locationPath << (absoluteLocationPath | relativeLocationPath)
        locationPath.setParseAction(
            lambda s, loc, toks: LocationPath(s, loc, toks, self.__getGraphRoot))

        self.__pathGrammer = locationPath

    def __substAlias(self, path):
        """Substitute aliases.

        Aliases are substituted at the first step of a relative location path.
        Further steps are not touched. An absolute location path is never
        substituted.
        """
        first, sep, tail = path.partition('/')
        if first:
            path = self.__aliases.get(first, first) + sep + tail
        return path

    def __getGraphRoot(self):
        """Get root node of package graph"""
        if self.__graph is None:
            # Try to load persisted graph. If the graph does not exist or does
            # not match it will be generated and saved.
            self.__graph = PkgGraphNode.init(".bob-tree.sqlite3", self.__cacheKey,
                self.getRootPackage)

        return self.__graph

    def __getGraphIter(self):
        """Get iterator that yields graph node and package together."""
        return GraphPackageIterator(self.__getGraphRoot(), self.getRootPackage())

    def __findResultNodes(self, node, result, valid, queryAll, stack=[]):
        if not queryAll: valid.discard(node)
        if node in result:
            if not queryAll: result.remove(node)
            yield (stack, node)
        for (name, child) in sorted((n, c.node) for (n, c) in node.items()
                                    if ((c.node in valid))):
            yield from self.__findResultNodes(child, result, valid,
                                              queryAll, stack + [name])

    def __findResultPackages(self, node, pkg, result, valid, queryAll):
        if not queryAll: valid.discard(node)
        nextPackages = { s.getPackage().getName() : s.getPackage()
            for s in pkg.getDirectDepSteps() }
        for s in pkg.getIndirectDepSteps():
            p = s.getPackage()
            nextPackages.setdefault(p.getName(), p)

        for (name, child) in sorted((n, c.node) for (n, c) in node.items() if c.node in valid):
            if child in result:
                if not queryAll: result.remove(child)
                yield nextPackages[name]
            yield from self.__findResultPackages(child, nextPackages[name], result, valid, queryAll)

    def __query(self, path):
        # replace aliases
        path = self.__substAlias(path)

        while path.endswith('/'): path = path[:-1]
        if path:
            try:
                path = self.__pathGrammer.parseString(path, True)
            except pyparsing.ParseBaseException as e:
                raise BobError("Invalid syntax: " + str(e),
                               help=markLocation(e.line, e.col))
            assert len(path) == 1
            assert isinstance(path[0], LocationPath)
            #print(path[0])
            return path[0].evalForward(self.__getGraphRoot())
        else:
            root = self.__getGraphRoot()
            return (set([root]), set([root]))

    def close(self):
        if self.__graph is not None:
            self.__graph.close()
            self.__graph = None

    def getAliases(self):
        return list(self.__aliases.keys())

    def getRootPackage(self):
        """Get virtual root package."""
        if self.__root is None:
            self.__root = self.__generator()
        return self.__root

    def queryTreePath(self, path, queryAll=False):
        """Execute query and return (stack, PkgGraphNode) tuples.

        Setting 'queryAll' to True will return all alternate paths to a result
        element instead of only the first one.
        """
        (nodes, valid) = self.__query(path)
        return self.__findResultNodes(self.__getGraphRoot(), nodes, valid, queryAll)

    def queryPackagePath(self, path, queryAll=False):
        """Execute query and return bob.input.Package objects.

        Setting 'queryAll' to True will return all alternate paths to a result
        package instead of only the first one.
        """
        (nodes, valid) = self.__query(path)
        return self.__findResultPackages(self.__getGraphRoot(), self.getRootPackage(), nodes, valid, queryAll)

    def walkPackagePath(self, path):
        """Legacy path walking.

        Does not support any advanced query features. Guarenteed to return only
        a single package. If any path element is not found a error is thrown.
        """
        # replace aliases
        path = self.__substAlias(path)

        # walk packages
        thisPackage = self.getRootPackage()
        steps = [ s for s in path.split("/") if s != "" ]
        if steps == []:
            raise BobError("'{}' is not a valid package path".format(path))
        for step in steps:
            nextPackages = { s.getPackage().getName() : s.getPackage()
                for s in thisPackage.getDirectDepSteps() }
            for s in thisPackage.getIndirectDepSteps():
                p = s.getPackage()
                nextPackages.setdefault(p.getName(), p)
            if step not in nextPackages:
                stack = thisPackage.getStack()
                raise BobError("Package '{}' not found under '{}'"
                                    .format(step, "/".join(stack) if stack != [''] else "/"))
            thisPackage = nextPackages[step]

        return thisPackage

    def getCacheKey(self):
        return self.__cacheKey



# Bob build tool
# Copyright (C) 2020  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..errors import BuildError
from ..stringparser import IfExpression
from ..tty import stepAction, INFO, TRACE
from ..utils import asHexStr, hashDirectory, emptyDirectory
from .scm import Scm, ScmAudit
import base64
import io
import os, os.path
import schema
import shutil
import stat
import tarfile

def copyTree(src, dst, invoker):
    """Recursively copy directory tree.

    The src and dst directories must already exist. The items in the source
    directory are copied to the destination only if it does not exist yet or if
    it is newer than the destination. The copy operation is aborted if the
    source and destination file types differ (file vs. directory vs. symlink).
    """

    changed = False
    try:
        for name in os.listdir(src):
            srcName = os.path.join(src, name)
            dstName = os.path.join(dst, name)

            srcStat = os.lstat(srcName)
            try:
                dstStat = os.lstat(dstName)
            except OSError:
                dstStat = None

            if dstStat is not None:
                if stat.S_IFMT(srcStat.st_mode ^ dstStat.st_mode) != 0:
                    invoker.fail("Copy failed: destination has different type:", srcName)

            if stat.S_ISLNK(srcStat.st_mode):
                linkTo = os.readlink(srcName)
                if dstStat is not None:
                    oldLink = os.readlink(dstName)
                    if linkTo == oldLink:
                        continue
                    os.unlink(dstName)
                os.symlink(linkTo, dstName)
                changed = True
            elif stat.S_ISDIR(srcStat.st_mode):
                if dstStat is None:
                    os.mkdir(dstName)
                    changed = True
                if copyTree(srcName, dstName, invoker):
                    shutil.copystat(srcName, dstName)
            else:
                if (dstStat is not None) and (srcStat.st_mtime_ns <= dstStat.st_mtime_ns):
                    continue
                shutil.copy2(srcName, dstName)
                changed = True
    except OSError as e:
        invoker.fail("Copy failed", str(e))

    return changed

def packTree(src):
    if not os.path.isdir(src):
        raise BuildError("Cannot import '{}': not a directory!".format(src))

    f = io.BytesIO()
    with tarfile.open(fileobj=f, mode="w:xz") as tar:
        tar.add(src, arcname=".")
    return base64.b85encode(f.getvalue()).decode('ascii')

def unpackTree(data, dest):
    f = io.BytesIO(base64.b85decode(data))
    with tarfile.open(fileobj=f, mode="r:xz") as tar:
        tar.extractall(dest)

class ImportScm(Scm):

    SCHEMA = schema.Schema({
        'scm' : 'import',
        'url' : str,
        schema.Optional('dir') : str,
        schema.Optional('if') : schema.Or(str, IfExpression),
        schema.Optional('prune') : bool,
    })

    def __init__(self, spec, overrides=[]):
        super().__init__(spec, overrides)
        self.__url = spec["url"]
        self.__dir = spec.get("dir", ".")
        self.__prune = spec.get("prune", False)
        self.__data = spec.get("__data")

    def getProperties(self):
        ret = super().getProperties()
        ret.update({
            'scm' : 'import',
            'url' : self.__url,
            'dir' : self.__dir,
            'prune' : self.__prune,
            '__data' : packTree(self.__url),
        })
        return ret

    async def invoke(self, invoker):
        dest = invoker.joinPath(self.__dir)
        os.makedirs(dest, exist_ok=True)
        if self.__prune: emptyDirectory(dest)
        if self.__data is None:
            if not os.path.isdir(self.__url):
                invoker.fail("Cannot import '{}': not a directory!".format(self.__url))
            copyTree(self.__url, dest, invoker)
        else:
            unpackTree(self.__data, dest)

    def asDigestScript(self):
        return self.__url

    def getDirectory(self):
        return self.__dir

    def isDeterministic(self):
        return False

    def hasLiveBuildId(self):
        return True

    async def predictLiveBuildId(self, step):
        with stepAction(step, "HASH", self.__url, (INFO, TRACE)) as a:
            return hashDirectory(self.__url)

    def calcLiveBuildId(self, workspacePath):
        return hashDirectory(os.path.join(workspacePath, self.__dir))

    def getAuditSpec(self):
        return ("import", self.__dir, {"url" : self.__url})

    def getLiveBuildIdSpec(self, workspacePath):
        return "#" + os.path.join(workspacePath, self.__dir)


class ImportAudit(ScmAudit):

    SCHEMA = schema.Schema({
        'type' : 'import',
        'dir' : str,
        'digest' : {
            'algorithm' : 'sha1',
            'value' : str
        },
        'url' : str,
    })

    async def _scanDir(self, workspace, dir, extra):
        self.__dir = dir
        self.__hash = asHexStr(hashDirectory(os.path.join(workspace, dir)))
        self.__url = extra.get("url")

    def _load(self, data):
        self.__dir = data["dir"]
        self.__hash = data["digest"]["value"]
        self.__url = data["url"]

    def dump(self):
        return {
            "type" : "import",
            "dir" : self.__dir,
            "digest" : {
                "algorithm" : "sha1",
                "value" : self.__hash
            },
            "url" : self.__url,
        }



# Bob build tool
# Copyright (C) 2016-2020, The BobBuildTool Contributors
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ...errors import ParseError
from ...input import RecipeSet
from ...utils import processDefines
from string import Formatter
import argparse
import os
import sys

from .builder import LocalBuilder
from .state import DevelopDirOracle

def doQueryPath(argv, bobRoot):
    # Configure the parser
    parser = argparse.ArgumentParser(prog="bob query-path",
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="""Query path information.

This command lists existing workspace directory names for packages given
on the command line. Output is formatted with a format string that can
contain placeholders
   {name}     package name
   {src}      checkout directory
   {build}    build directory
   {dist}     package directory
The default format is '{name}<tab>{dist}'.

If a directory does not exist for a step (because that step has never
been executed or does not exist), the line is omitted.
""")
    parser.add_argument('packages', metavar='PACKAGE', type=str, nargs='+',
        help="(Sub-)package to query")
    parser.add_argument('-f', help='Output format string', default='{name}\t{dist}', metavar='FORMAT')
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    parser.add_argument('-q', dest="quiet", action="store_true",
        help="Be quiet in case of errors")
    parser.add_argument('--fail', action="store_true",
        help="Return a non-zero error code in case of errors")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox', help="Disable sandboxing")
    parser.set_defaults(sandbox=None)

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--develop', action='store_true',  dest='dev', help="Use developer mode", default=True)
    group.add_argument('--release', action='store_false', dest='dev', help="Use release mode")

    # Parse args
    args = parser.parse_args(argv)
    if args.sandbox == None:
        args.sandbox = not args.dev

    defines = processDefines(args.defines)

    # Process the recipes
    recipes = RecipeSet()
    recipes.defineHook('releaseNameFormatter', LocalBuilder.releaseNameFormatter)
    recipes.defineHook('developNameFormatter', LocalBuilder.developNameFormatter)
    recipes.defineHook('developNamePersister', None)
    recipes.setConfigFiles(args.configFile)
    recipes.parse()

    # State variables in a class
    class State:
        def __init__(self):
            self.packageText = ''
            self.showPackage = True
            self.failedSteps = []
        def appendText(self, what):
            self.packageText += what
        def appendStep(self, step):
            dir = step.getWorkspacePath()
            if step.isValid() and (dir is not None) and os.path.isdir(dir):
                self.packageText += dir
            else:
                self.showPackage = False
                self.failedSteps.append(step)
        def print(self):
            if (self.showPackage):
                print(self.packageText)
            else:
                if not args.quiet:
                    packageName = self.failedSteps[0].getPackage().getName()
                    if len(self.failedSteps) is 1:
                        print("Directory for {{{}}} step of package {} not present.".format(
                            self.failedSteps[0].getLabel(), packageName), file=sys.stderr)
                    else:
                        labelList = ', '.join([step.getLabel() for step in self.failedSteps])
                        print("Directories for {{{}}} steps of package {} not present.".format(
                            labelList, packageName), file=sys.stderr)
                if args.fail:
                    sys.exit(1)

    if args.dev:
        # Develop names are stable. All we need to do is to replicate build's algorithm,
        # and when we produce a name, check whether it exists.
        nameFormatter = recipes.getHook('developNameFormatter')
        developPersister = DevelopDirOracle(nameFormatter, recipes.getHook('developNamePersister'))
        nameFormatter = developPersister.getFormatter()
    else:
        # Release names are taken from persistence.
        nameFormatter = LocalBuilder.releaseNameInterrogator
    nameFormatter = LocalBuilder.makeRunnable(nameFormatter)

    # Find roots
    packages = recipes.generatePackages(nameFormatter, defines, args.sandbox)
    if args.dev: developPersister.prime(packages)

    matched = False
    # Loop through packages
    for p in args.packages:
        # Format this package.
        # Only show the package if all of the requested directory names are present
        for package in packages.queryPackagePath(p):
            matched = True
            state = State()
            for (text, var, spec, conversion) in Formatter().parse(args.f):
                state.appendText(text)
                if var is None:
                    pass
                elif var == 'name':
                    state.appendText("/".join(package.getStack()))
                elif var == 'src':
                    state.appendStep(package.getCheckoutStep())
                elif var == 'build':
                    state.appendStep(package.getBuildStep())
                elif var == 'dist':
                    state.appendStep(package.getPackageStep())
                else:
                    raise ParseError("Unknown field '{" + var + "}'")

            # Show
            state.print()

    if not matched:
        if not args.quiet:
            print("Your query matched no packages. Naptime!", file=sys.stderr)
        if args.fail:
            sys.exit(1)
